(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.HXS = {}));
})(this, (function (exports) { 'use strict';

    var HELP_SYMBOL = Symbol('hxs-help-symbol');
    var CONSTRUCTOR_SYMBOL = Symbol('hxs-constructor-symbol');
    var BASE_SYMBOL = Symbol('hxs-base-symbol');
    var PROMISE_SYMBOL = Symbol('hxs-promise-symbol');
    exports.Utils = void 0;

    (function (Utils) {
      /**
       * A stable in-place sorting function.
       */
      Utils.sort = function (array, compare) {
        var t;

        for (var i = 0; i < array.length - 1; i++) {
          for (var j = i + 1; j < array.length; j++) {
            if (compare(array[i], array[j]) > 0) {
              t = array[i];
              array[i] = array[j];
              array[j] = t;
            }
          }
        }
      };
      /**
       * Remove specific element(s) from the array.
       * (Default count: `array.length - start`)
       */


      Utils.removeElements = function (array, start, count) {
        if (count === void 0) {
          count = array.length - start;
        }

        for (var i = start; i < array.length - count; i++) {
          array[i] = array[i + count];
        }

        array.length -= count;
      };

      Utils.normalizeIndex = function (index, arraySize, referrer, context, allowEnd) {
        if (allowEnd === void 0) {
          allowEnd = false;
        }

        var normalizedIndex = index < 0 ? arraySize + index : index;

        if (normalizedIndex > arraySize || normalizedIndex < 0 || !allowEnd && normalizedIndex === arraySize) {
          var message = "index(" + index + ") out of range(array size: " + arraySize + ")";
          Utils.raise(RangeError, message, referrer, context);
        }

        return normalizedIndex;
      };

      Utils.formatFrameString = function (referrer, context) {
        return "    at " + context.source + " (Ln " + referrer.line + ", Col " + referrer.column + ")";
      };
      /**
       * Throw a specific error with environment info appended.
       */


      Utils.raise = function (constructor, message, referrer, context) {
        var errorMessage = message;

        if (referrer && context) {
          var currentFrame = Utils.formatFrameString(referrer, context);
          var stack = context.stack;

          if (stack[stack.length - 1] !== currentFrame) {
            errorMessage += '\n' + currentFrame;
          }

          for (var i = stack.length - 1; i >= 0; i--) {
            errorMessage += '\n' + stack[i];
          }
        }

        throw new constructor(errorMessage);
      };

      Utils.parseNumber = function (node, context) {
        var result;

        switch (node.suffix) {
          case 'D':
          case '':
            {
              result = +node.value;
              break;
            }

          case 'B':
            {
              result = Number.parseInt(node.value, 2);
              break;
            }

          case 'O':
            {
              result = Number.parseInt(node.value, 8);
              break;
            }

          case 'H':
            {
              result = Number.parseInt(node.value, 16);
              break;
            }

          default:
            {
              Utils.raise(SyntaxError, 'unrecognized number', node, context);
            }
        }

        if (result !== result) {
          // NaN
          Utils.raise(SyntaxError, 'invalid number', node, context);
        }

        return result;
      };

      Utils.isDict = function (value) {
        return Object.prototype.toString.call(value) === '[object Object]';
      };

      Utils.createDict = function (dict) {
        return Object.assign(Object.create(null), dict);
      };

      Utils.createValueNode = function (value, referrer) {
        return {
          type: 'value',
          line: referrer.line,
          column: referrer.column,
          offset: referrer.offset,
          value: value
        };
      };

      Utils.createRawArray = function (array, referrer) {
        var COMMA_NODE = {
          type: 'symbol',
          value: ',',
          line: referrer.line,
          column: referrer.column,
          offset: referrer.offset
        };
        var rawArray = [];

        for (var i = 0; i < array.length; i++) {
          if (i) {
            rawArray.push(COMMA_NODE);
          }

          rawArray.push(Utils.createValueNode(array[i], referrer));
        }

        return rawArray;
      };

      Utils.replaceBuffer = function (buffer, start, width, replacement) {
        buffer[start] = replacement;

        if (width > 1) {
          Utils.removeElements(buffer, start + 1, width - 1);
        }
      };
      /**
       * Excute the callback with the temporary variables
       * defined in context store and recover conflicting
       * variables after execution.
       */


      Utils.injectTemp = function (store, variables, callback) {
        var copy = Object.create(null);

        for (var key in variables) {
          if (key in store) {
            copy[key] = store[key];
          }

          store[key] = variables[key];
        }

        callback();

        for (var _key in variables) {
          if (_key in copy) {
            store[_key] = copy[_key];
          } else {
            delete store[_key];
          }
        }
      };

      Utils.extendContext = function (context, store, stack) {
        return {
          store: store,
          exports: context.exports,
          resolvedModules: context.resolvedModules,
          source: context.source,
          basePath: context.basePath,
          stack: stack
        };
      };

      Utils.injectHelp = function (helpInfo, target) {
        target[HELP_SYMBOL] = helpInfo;
        return target;
      };

      Utils.toString = function (value, previewSize, previewIndent) {
        if (previewSize === void 0) {
          previewSize = 10;
        }

        if (previewIndent === void 0) {
          previewIndent = '  ';
        }

        var type = typeof value;

        if (type === 'string') {
          return value;
        } else if (type === 'function') {
          return '<function>';
        } else if (Array.isArray(value)) {
          if (previewSize <= 0) {
            return '<array>';
          }

          var result = "(size: " + value.length + ") [";

          if (value.length > 0) {
            result += '\n' + previewIndent;
            result += value.slice(0, previewSize).map(function (v) {
              return Utils.toDisplay(v, 0) + ',\n';
            }).join(previewIndent);
          }

          if (value.length > previewSize) {
            result += previewIndent + '...\n';
          }

          result += ']';
          return result;
        } else if (Utils.isDict(value)) {
          if (previewSize <= 0) {
            return '<dict>';
          }

          var entries = Object.entries(value);

          var _result = "(size: " + entries.length + ") {";

          if (entries.length > 0) {
            _result += '\n' + previewIndent;
            _result += entries.slice(0, previewSize).map(function (entry) {
              return Utils.toDisplay(entry[0], 0) + ' -> ' + Utils.toDisplay(entry[1], 0) + ',\n';
            }).join(previewIndent);

            if (entries.length > previewSize) {
              _result += previewIndent + '...\n';
            }
          }

          _result += '}';
          return _result;
        } else {
          return String(value);
        }
      };

      Utils.toDisplay = function (value, previewSize, previewIndent) {
        if (typeof value !== 'string') {
          return Utils.toString(value, previewSize, previewIndent);
        }

        if (!value.includes("'")) {
          return "'" + value + "'";
        } else if (!value.includes('"')) {
          return '"' + value + '"';
        } else if (!value.includes('`')) {
          return '`' + value + '`';
        } else {
          return "'" + value.replace(/'/g, "\\'") + "'";
        }
      };

      Utils.filterValue = function (value) {
        if (value === undefined) {
          return null;
        } else if (value && typeof value === 'object') {
          if (Utils.isDict(value)) {
            return value;
          } else {
            return String(value);
          }
        } else {
          return value;
        }
      };
    })(exports.Utils || (exports.Utils = {}));

    var getOperatorSymbol = function getOperatorSymbol(node) {
      return node.type === 'symbol' ? node.value : node.start;
    };

    /**
     * Evalute a single node.
     */

    var evalNode = function evalNode(node, context) {
      switch (node.type) {
        case 'value':
          {
            return node.value;
          }

        case 'glob':
          {
            return node.value.slice(1, -1);
          }

        case 'number':
          {
            return exports.Utils.parseNumber(node, context);
          }

        case 'span':
          {
            return evalExpression(node.body, context);
          }

        case 'word':
          {
            var name = node.value;
            var store = context.store;

            if (!(name in store)) {
              exports.Utils.raise(ReferenceError, exports.Utils.toDisplay(name) + " is not defined", node, context);
            }

            return store[name];
          }

        default:
          {
            exports.Utils.raise(SyntaxError, 'unrecognized syntax', node, context);
            return null; // for type checking (ts v4.3.4)
          }
      }
    };

    var compileExpression = function compileExpression(buffer, context) {
      var operatorNodes = []; // find all operators (and spans)

      for (var i = 0; i < buffer.length; i++) {
        var node = buffer[i];

        if (node.type === 'symbol') {
          operatorNodes.push(node);

          if (!operatorPriorities.has(node.value)) {
            exports.Utils.raise(SyntaxError, "unknown operator " + exports.Utils.toDisplay(node.value), node, context);
          }
        } else if (node.type === 'span') {
          if (operatorPriorities.has(node.start)) {
            operatorNodes.push(node);
          }
        }
      } // sort operators in order of priority (high->low)


      exports.Utils.sort(operatorNodes, function (nodeA, nodeB) {
        var priorityA = operatorPriorities.get(getOperatorSymbol(nodeA));
        var priorityB = operatorPriorities.get(getOperatorSymbol(nodeB));
        return priorityA - priorityB;
      });
      var p = 0;
      var t;

      while (p < operatorNodes.length) {
        var symbol = getOperatorSymbol(operatorNodes[p]);

        for (var q = p + 1; q <= operatorNodes.length; q++) {
          if (q < operatorNodes.length && getOperatorSymbol(operatorNodes[q]) === symbol) {
            continue;
          }

          if (operatorLtr.get(symbol)) {
            p = q;
            break;
          }

          var l = p;
          var r = q - 1;

          while (l < r) {
            t = operatorNodes[l];
            operatorNodes[l] = operatorNodes[r];
            operatorNodes[r] = t;
            l++;
            r--;
          }

          p = q;
        }
      }

      return {
        buffer: buffer,
        operatorNodes: operatorNodes
      };
    };
    var evalCompiledExpression = function evalCompiledExpression(compiledExpression, context, copyBuffers, optimizeOperators) {
      var operatorNodes = compiledExpression.operatorNodes,
          _buffer = compiledExpression.buffer;
      var buffer = copyBuffers ? _buffer.slice() : _buffer;
      var extraOperatorIndices = null;

      if (optimizeOperators) {
        extraOperatorIndices = new Set();
      } // execute operators


      for (var i = 0; i < operatorNodes.length; i++) {
        var operatorNode = operatorNodes[i];
        var index = buffer.indexOf(operatorNode);

        if (~index) {
          // index !== -1
          var handler = operatorHandlers.get(operatorNode.type === 'symbol' ? operatorNode.value : operatorNode.start);
          handler(buffer, index, context);
        } else if (optimizeOperators) {
          extraOperatorIndices.add(i);
        }
      } // remove extra operators


      if (extraOperatorIndices) {
        var optimizedOperators = operatorNodes.filter(function (_, i) {
          return !extraOperatorIndices.has(i);
        });

        if (optimizedOperators.length < operatorNodes.length) {
          for (var _i = 0; _i < optimizedOperators.length; _i++) {
            operatorNodes[_i] = optimizedOperators[_i];
          }

          operatorNodes.length = optimizedOperators.length;
        }
      } // check result


      if (buffer.length > 1) {
        exports.Utils.raise(SyntaxError, 'unrecognized syntax', buffer[1], context);
        return null; // for type checking
      } else if (buffer.length === 0) {
        return null; // for type checking
      } else {
        return evalNode(buffer[0], context);
      }
    };
    /**
     * Evalute an expression.
     */

    var evalExpression = function evalExpression(nodes, context, begin, end) {
      if (begin === void 0) {
        begin = 0;
      }

      if (end === void 0) {
        end = nodes.length;
      }

      if (begin >= end) {
        return null;
      }

      var buffer = nodes.slice(begin, end);
      return evalCompiledExpression(compileExpression(buffer, context), context, false, false);
    };

    var compileNodes = function compileNodes(nodes, context, begin, end) {
      if (begin === void 0) {
        begin = 0;
      }

      if (end === void 0) {
        end = nodes.length;
      }

      var compiledExpressions = [];

      if (begin >= end) {
        return {
          compiledExpressions: compiledExpressions,
          endsWithSemicolon: true
        };
      }

      var left = begin;
      var right;

      for (right = 0; right <= end; right++) {
        if (right < end) {
          var node = nodes[right];

          if (node.type !== 'symbol' || node.value !== ';') {
            continue;
          }
        }

        var buffer = nodes.slice(left, right);
        compiledExpressions.push(compileExpression(buffer, context));

        if (right < end - 1) {
          left = right + 1;
        } else {
          break;
        }
      }

      var endsWithSemicolon = right < end;
      return {
        compiledExpressions: compiledExpressions,
        endsWithSemicolon: endsWithSemicolon
      };
    };
    var evalCompiledNodes = function evalCompiledNodes(compiledNodes, context, copyBuffers, optimizeOperators) {
      var compiledExpressions = compiledNodes.compiledExpressions,
          endsWithSemicolon = compiledNodes.endsWithSemicolon;
      var lastValue = null;

      for (var i = 0; i < compiledExpressions.length; i++) {
        lastValue = evalCompiledExpression(compiledExpressions[i], context, copyBuffers, optimizeOperators);
        context.store._ = lastValue;
      }

      return endsWithSemicolon ? null : lastValue;
    };
    /**
     * Evaluate the given nodes.
     */

    var evalNodes = function evalNodes(nodes, context, begin, end) {
      if (begin === void 0) {
        begin = 0;
      }

      if (end === void 0) {
        end = nodes.length;
      }

      return evalCompiledNodes(compileNodes(nodes, context, begin, end), context, false, false);
    };

    /**
     * Parse single argument declaration.
     */

    var parseArg = function parseArg(rawArgs, begin, end, context, allowRequired) {
      var firstNode = rawArgs[begin];
      var secondNode = rawArgs[begin + 1];

      if (!firstNode || firstNode.type !== 'word') {
        exports.Utils.raise(SyntaxError, 'expect a word as argument name', firstNode, context);
      }

      if (end - begin === 1) {
        if (!allowRequired) {
          exports.Utils.raise(SyntaxError, 'required argument must be before optional ones', firstNode, context);
        }

        return {
          name: firstNode.value,
          required: true,
          rest: false,
          "default": null
        };
      } else if (end - begin === 2) {
        if (secondNode.type !== 'symbol' || secondNode.value !== '...' && secondNode.value !== '?') {
          exports.Utils.raise(SyntaxError, 'invalid argument declaration', firstNode, context);
        }

        return {
          name: firstNode.value,
          required: false,
          rest: secondNode.value === '...',
          "default": null
        };
      } else {
        if (secondNode.type !== 'symbol' || secondNode.value !== '=') {
          exports.Utils.raise(SyntaxError, 'invalid argument declaration', firstNode, context);
        }

        return {
          name: firstNode.value,
          required: false,
          rest: false,
          "default": compileExpression(rawArgs.slice(begin + 2, end), context)
        };
      }
    };
    /**
     * Parse argument list.
     */

    var parseArgList = function parseArgList(rawArgList, context) {
      if (rawArgList.length === 0) {
        return {
          args: [],
          requiredCount: 0,
          restArg: null
        };
      }

      var args = [];
      var allowRequired = true;
      var requiredCount = 0;
      var l = 0;
      var argDef;
      var restArg = null;

      for (var r = 0; r <= rawArgList.length; r++) {
        if (r < rawArgList.length) {
          var node = rawArgList[r];

          if (node.type !== 'symbol' || node.value !== ',') {
            continue;
          }
        }

        if (restArg !== null) {
          exports.Utils.raise(SyntaxError, 'rest argument must be the last one', rawArgList[l], context);
        }

        argDef = parseArg(rawArgList, l, r, context, allowRequired);

        if (argDef.rest) {
          restArg = argDef.name;
        } else {
          args.push(argDef);

          if (argDef.required) {
            requiredCount++;
          } else {
            allowRequired = false;
          }
        }

        if (r < rawArgList.length - 1) {
          l = r + 1;
        } else {
          break;
        }
      }

      return {
        args: args,
        requiredCount: requiredCount,
        restArg: restArg
      };
    };
    var isInvocable = function isInvocable(target) {
      return typeof target === 'function' || exports.Utils.isDict(target) && '__invoke' in target;
    };
    var invoke = function invoke(target, rawArgs, referrer, context, thisArg) {
      context.stack.push(exports.Utils.formatFrameString(referrer, context));

      if (typeof target === 'function') {
        var returnValue = target(rawArgs, referrer, context, thisArg);
        context.stack.pop();
        return returnValue;
      } else if (exports.Utils.isDict(target) && '__invoke' in target) {
        if (typeof target.__invoke !== 'function') {
          exports.Utils.raise(TypeError, 'expect `__invoke` to be a function', referrer, context);
        }

        var _returnValue = target.__invoke(rawArgs, referrer, context, thisArg);

        context.stack.pop();
        return _returnValue;
      } else {
        exports.Utils.raise(TypeError, 'expect a function or dict with proper `__invoke`', referrer, context);
        return null; // for type checking
      }
    };

    /**
     * Evalute an list.
     */

    var evalList = function evalList(nodes, context, begin, end) {
      if (begin === void 0) {
        begin = 0;
      }

      if (end === void 0) {
        end = nodes.length;
      }

      var result = [];

      if (begin >= end) {
        return result;
      }

      var left = begin;

      for (var right = 0; right <= end; right++) {
        if (right < end) {
          var node = nodes[right];

          if (node.type !== 'symbol' || node.value !== ',') {
            continue;
          }
        }

        var firstNode = nodes[left];

        if (firstNode.type === 'symbol' && firstNode.value === '...') {
          var values = evalExpression(nodes, context, left + 1, right);

          if (!Array.isArray(values)) {
            exports.Utils.raise(TypeError, 'expect an array following', firstNode, context);
          }

          for (var i = 0; i < values.length; i++) {
            result.push(values[i]);
          }
        } else {
          result.push(evalExpression(nodes, context, left, right));
        }

        if (right < end - 1) {
          left = right + 1;
        } else {
          break;
        }
      }

      return result;
    };

    var createFunctionHandler = function createFunctionHandler(minArgCount, maxArgCount, callback) {
      return function (rawArgs, referrer, context, thisArg) {
        var args = evalList(rawArgs, context);

        if (args.length < minArgCount) {
          exports.Utils.raise(TypeError, 'too few arguments', referrer, context);
        } else if (args.length > maxArgCount) {
          exports.Utils.raise(TypeError, 'too many arguments', referrer, context);
        }

        return callback(args, referrer, context, thisArg);
      };
    };

    /**
     * Create an expression from source code.
     */

    var createInlineExpression = function createInlineExpression(buffer, index, context) {
      var argSpan = buffer[index - 1];
      var bodySpan = buffer[index + 1];

      if (index + 1 >= buffer.length || argSpan.type !== 'span' || argSpan.start !== '(' || bodySpan.type !== 'span' || bodySpan.start !== '(') {
        exports.Utils.raise(SyntaxError, 'invalid expression declaration', buffer[index], context);
      }

      var argList = parseArgList(argSpan.body, context);
      var body = bodySpan.body;
      var compiledBody = compileNodes(body, context);
      return createFunctionHandler(argList.requiredCount, Infinity, function (args, referrer, _context, thisArg) {
        var scopeStore = exports.Utils.createDict(context.store);
        scopeStore._ = null;
        scopeStore["this"] = thisArg;
        scopeStore.arguments = args;
        var scopeContext = exports.Utils.extendContext(context, scopeStore, _context.stack.slice());
        var argDefinitions = argList.args;

        for (var i = 0; i < argDefinitions.length; i++) {
          var argName = argDefinitions[i].name;

          if (i < args.length) {
            scopeStore[argName] = args[i];
          } else {
            var defaultValue = argDefinitions[i]["default"];

            if (defaultValue !== null) {
              scopeStore[argName] = evalCompiledExpression(defaultValue, scopeContext, true, true);
            } else {
              scopeStore[argName] = null;
            }
          }
        }

        if (argList.restArg !== null) {
          scopeStore[argList.restArg] = args.slice(argDefinitions.length);
        } // rewrite


        scopeStore._ = null;
        scopeStore["this"] = thisArg;
        scopeStore.arguments = args;
        return evalCompiledNodes(compiledBody, scopeContext, true, true);
      });
    };
    var inlineExpressionHandler = function inlineExpressionHandler(buffer, index, context) {
      var func = createInlineExpression(buffer, index, context);
      var valueNode = exports.Utils.createValueNode(func, buffer[index]);
      exports.Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
    };

    /**
     * Create a function from source code.
     */

    var createInlineFunction = function createInlineFunction(buffer, index, context) {
      var argSpan = buffer[index + 1];
      var bodySpan = buffer[index + 2];

      if (index + 2 >= buffer.length || argSpan.type !== 'span' || argSpan.start !== '(' || bodySpan.type !== 'span' || bodySpan.start !== '{') {
        exports.Utils.raise(SyntaxError, 'invalid function declaration', buffer[index], context);
      }

      var argList = parseArgList(argSpan.body, context);
      var body = bodySpan.body;
      var compiledBody = compileNodes(body, context);
      return createFunctionHandler(argList.requiredCount, Infinity, function (args, referrer, _context, thisArg) {
        var scopeStore = exports.Utils.createDict(context.store);
        scopeStore._ = null;
        scopeStore["this"] = thisArg;
        scopeStore.arguments = args;
        var scopeContext = exports.Utils.extendContext(context, scopeStore, _context.stack.slice());
        var argDefinitions = argList.args;

        for (var i = 0; i < argDefinitions.length; i++) {
          var argName = argDefinitions[i].name;

          if (i < args.length) {
            scopeStore[argName] = args[i];
          } else {
            var defaultValue = argDefinitions[i]["default"];

            if (defaultValue !== null) {
              scopeStore[argName] = evalCompiledExpression(defaultValue, scopeContext, true, true);
            } else {
              scopeStore[argName] = null;
            }
          }
        }

        if (argList.restArg !== null) {
          scopeStore[argList.restArg] = args.slice(argDefinitions.length);
        } // rewrite


        scopeStore._ = null;
        scopeStore["this"] = thisArg;
        scopeStore.arguments = args;
        var RETURN_FLAG = Symbol('hxs-return-flag');
        var returnValue = null;
        scopeStore["return"] = createFunctionHandler(0, 1, function (args) {
          if (args.length) {
            returnValue = args[0];
          }

          throw RETURN_FLAG;
        });
        scopeStore.forward = createFunctionHandler(1, 1, function (args, _referrer) {
          var names = args[0];

          if (!Array.isArray(names)) {
            exports.Utils.raise(TypeError, 'expect an array of strings', _referrer, context);
          }

          for (var _i = 0; _i < names.length; _i++) {
            var name = names[_i];

            if (typeof name !== 'string') {
              exports.Utils.raise(TypeError, 'expect an array of strings', _referrer, context);
            }

            if (name in scopeStore) {
              context.store[name] = scopeStore[name];
            } else {
              exports.Utils.raise(ReferenceError, exports.Utils.toDisplay(name) + " is not defined", _referrer, context);
            }
          }

          return null;
        });

        try {
          evalCompiledNodes(compiledBody, scopeContext, true, true);
        } catch (err) {
          if (err !== RETURN_FLAG) {
            throw err;
          }
        }

        return returnValue;
      });
    };
    var inlineFunctionHandler = function inlineFunctionHandler(buffer, index, context) {
      var secondNode = buffer[index + 1];

      if (secondNode.type === 'word') {
        // named
        var name = secondNode.value;
        var func = createInlineFunction(buffer, index + 1, context);
        context.store[name] = func;
        var valueNode = exports.Utils.createValueNode(func, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index, 4, valueNode);
      } else {
        // anonymous
        var _func = createInlineFunction(buffer, index, context);

        var _valueNode = exports.Utils.createValueNode(_func, buffer[index]);

        exports.Utils.replaceBuffer(buffer, index, 3, _valueNode);
      }
    };

    var getKeysOf = function getKeysOf(dict, referrer, context) {
      var keys = isInvocable(dict.__keys) ? invoke(dict.__keys, [], referrer, context, null) : Object.keys(dict);

      if (!Array.isArray(keys)) {
        exports.Utils.raise(TypeError, '`__keys` should return an array', referrer, context);
      }

      return keys;
    };
    var hasProperty = function hasProperty(dict, key, keyReferrer, operationReferrer, context) {
      if (isInvocable(dict.__has)) {
        return invoke(dict.__has, [exports.Utils.createValueNode(key, keyReferrer)], operationReferrer, context, null);
      } else {
        if (typeof key !== 'string') {
          exports.Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
        }

        var keys = getKeysOf(dict, operationReferrer, context);
        return keys.includes(key);
      }
    };
    var removeProperty = function removeProperty(target, key, keyReferrer, operationReferrer, context) {
      if (Array.isArray(target)) {
        if (typeof key !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as index', keyReferrer, context);
        }

        var normalizedIndex = exports.Utils.normalizeIndex(key, target.length, keyReferrer, context);
        exports.Utils.removeElements(target, normalizedIndex, 1);
      } else if (exports.Utils.isDict(target)) {
        if (isInvocable(target.__remove)) {
          invoke(target.__remove, [exports.Utils.createValueNode(key, keyReferrer)], operationReferrer, context, null);
        } else {
          if (typeof key !== 'string') {
            exports.Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
          }

          if (key in target) {
            // Don't use `hasProperty` here!
            delete target[key];
          }
        }
      } else {
        exports.Utils.raise(TypeError, 'invalid property removal', operationReferrer, context);
      }
    };
    var getProperty = function getProperty(target, key, allowMagic, keyReferrer, operationReferrer, context) {
      if (Array.isArray(target) || typeof target === 'string') {
        if (typeof key !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as index', keyReferrer, context);
        }

        var normalizedIndex = exports.Utils.normalizeIndex(key, target.length, keyReferrer, context);
        return target[normalizedIndex];
      } else if (exports.Utils.isDict(target)) {
        if (allowMagic && isInvocable(target.__get)) {
          return invoke(target.__get, [exports.Utils.createValueNode(key, keyReferrer)], operationReferrer, context, null);
        } else {
          if (typeof key !== 'string') {
            exports.Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
          }

          if (key in target) {
            // Don't use `hasProperty` here!
            return target[key];
          } else {
            return null;
          }
        }
      } else {
        exports.Utils.raise(TypeError, 'invalid property access', operationReferrer, context);
        return null; // for type checking
      }
    };
    var setProperty = function setProperty(target, key, value, allowMagic, keyReferrer, operationReferrer, context) {
      if (Array.isArray(target)) {
        if (typeof key !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as index', keyReferrer, context);
        }

        var normalizedIndex = exports.Utils.normalizeIndex(key, target.length, keyReferrer, context);
        target[normalizedIndex] = value;
      } else if (exports.Utils.isDict(target)) {
        if (allowMagic && isInvocable(target.__set)) {
          invoke(target.__set, exports.Utils.createRawArray([key, value], operationReferrer), operationReferrer, context, target);
        } else {
          if (typeof key !== 'string') {
            exports.Utils.raise(TypeError, 'expect a string as key', keyReferrer, context);
          }

          target[key] = value;
        }
      } else {
        exports.Utils.raise(TypeError, 'invalid property assignment', operationReferrer, context);
      }
    };
    var createClass = function createClass(description, referrer, context) {
      var constructor = exports.Utils.createDict({
        __invoke: function __invoke(_rawArgs, _referrer, _context, _thisArg) {
          var object = _thisArg;

          if (!exports.Utils.isDict(object)) {
            // no prototype
            if (object !== null) {
              exports.Utils.raise(TypeError, 'expect only null or dict as this arg for class constructor', referrer, context);
            }

            object = Object.create(null);
            object[CONSTRUCTOR_SYMBOL] = constructor;
          }

          var _loop = function _loop(key) {
            if (key in object) {
              return "continue";
            }

            var value = description[key];

            if (typeof value === 'function') {
              object[key] = function (rawArgs, referrer, context) {
                return value(rawArgs, referrer, context, object);
              };
            } else {
              object[key] = value;
            }
          };

          for (var key in description) {
            var _ret = _loop(key);

            if (_ret === "continue") continue;
          }

          if (description.__init) {
            invoke(description.__init, _rawArgs, _referrer, _context, object);
          }

          return object;
        }
      });
      return constructor;
    };
    var getConstructorOf = function getConstructorOf(dict) {
      return CONSTRUCTOR_SYMBOL in dict ? dict[CONSTRUCTOR_SYMBOL] : null;
    };
    var isInstanceOf = function isInstanceOf(constructor, dict) {
      var dictConstructor = getConstructorOf(dict);

      if (constructor === dictConstructor) {
        return true;
      }

      while (dictConstructor) {
        if (exports.Utils.isDict(dictConstructor) && BASE_SYMBOL in dictConstructor) {
          dictConstructor = dictConstructor[BASE_SYMBOL];

          if (constructor === dictConstructor) {
            return true;
          }
        } else {
          break;
        }
      }

      return false;
    };

    /**
     * Execute and replace the given node of the buffer.
     */

    var evalBufferNode = function evalBufferNode(buffer, index, referrer, context) {
      if (index < 0 || index >= buffer.length) {
        exports.Utils.raise(SyntaxError, 'invalid operation', referrer, context);
      }

      return evalNode(buffer[index], context);
    };

    var evalBinaryOperation = function evalBinaryOperation(a, b, magicName, typeA, typeB, defaultHandler, buffer, index, context) {
      if (magicName) {
        if (exports.Utils.isDict(a) && magicName in a) {
          var magicMethod = a[magicName];

          if (!isInvocable(magicMethod)) {
            exports.Utils.raise(TypeError, "expect an invocable as " + magicName, buffer[index - 1], context);
          }

          return invoke(magicMethod, exports.Utils.createRawArray([b, false], buffer[index]), buffer[index], context, null);
        } else if (exports.Utils.isDict(b) && magicName in b) {
          var _magicMethod = b[magicName];

          if (!isInvocable(_magicMethod)) {
            exports.Utils.raise(TypeError, "expect an invocable as " + magicName, buffer[index + 1], context);
          }

          return invoke(_magicMethod, exports.Utils.createRawArray([a, true], buffer[index]), buffer[index], context, null);
        }
      }

      if (typeA && typeof a !== typeA) {
        exports.Utils.raise(TypeError, "expect a " + typeA, buffer[index - 1], context);
      }

      if (typeB && typeof b !== typeB) {
        exports.Utils.raise(TypeError, "expect a " + typeB, buffer[index + 1], context);
      }

      return defaultHandler(a, b);
    };
    var evalUnaryOperation = function evalUnaryOperation(value, magicName, type, defaultHandler, buffer, index, context) {
      if (magicName && exports.Utils.isDict(value) && magicName in value) {
        var magicMethod = value[magicName];

        if (!isInvocable(magicMethod)) {
          exports.Utils.raise(TypeError, "expect an invocable as " + magicName, buffer[index - 1], context);
        }

        return invoke(magicMethod, [exports.Utils.createValueNode(value, buffer[index])], buffer[index], context, null);
      } else {
        if (type && typeof value !== type) {
          exports.Utils.raise(TypeError, "expect a " + type, buffer[index - 1], context);
        }

        return defaultHandler(value);
      }
    };
    var createBinaryOperator = function createBinaryOperator(magicName, typeA, typeB, defaultHandler) {
      return function (buffer, index, context) {
        var a = evalBufferNode(buffer, index - 1, buffer[index], context);
        var b = evalBufferNode(buffer, index + 1, buffer[index], context);
        var valueNode = exports.Utils.createValueNode(evalBinaryOperation(a, b, magicName, typeA, typeB, defaultHandler, buffer, index, context), buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
      };
    };
    var createUnaryOperator = function createUnaryOperator(magicName, type, defaultHandler) {
      return function (buffer, index, context) {
        var value = evalBufferNode(buffer, index + 1, buffer[index], context);
        var valueNode = exports.Utils.createValueNode(evalUnaryOperation(value, magicName, type, defaultHandler, buffer, index, context), buffer[index]);
        exports.Utils.replaceBuffer(buffer, index, 2, valueNode);
      };
    };
    var createAdditionalAssignmentOperator = function createAdditionalAssignmentOperator(magicName, typeA, typeB, defaultHandler) {
      return function (buffer, index, context) {
        var operationReferrer = buffer[index];

        if (index === 0) {
          exports.Utils.raise(SyntaxError, 'expect a variable preceding', operationReferrer, context);
        }

        var precedingNode = buffer[index - 1];
        var target = null;
        var key = null;
        var allowMagic = false;

        if (precedingNode.type === 'word') {
          // a = b or a.b = c
          if (index === 1) {
            // a = b
            var name = precedingNode.value;

            if (!(name in context.store)) {
              exports.Utils.raise(SyntaxError, exports.Utils.toDisplay(name) + " is not defined", precedingNode, context);
            }

            target = context.store;
            key = precedingNode.value;
          } else {
            // a.b = c
            if (index === 2 || buffer[index - 2].type !== 'symbol' || buffer[index - 2].value !== '.') {
              exports.Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
            }

            target = evalExpression(buffer, context, 0, index - 2);

            if (!exports.Utils.isDict(target)) {
              exports.Utils.raise(TypeError, 'expect a dict as assignment target', buffer[0], context);
            }

            key = precedingNode.value;
          }
        } else if (precedingNode.type === 'span' && precedingNode.start === '[') {
          // a[b] = c
          target = evalExpression(buffer, context, 0, index - 1);
          key = evalExpression(precedingNode.body, context);
          allowMagic = true;
        } else {
          exports.Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
        }

        var a = getProperty(target, key, allowMagic, precedingNode, operationReferrer, context);
        var b = evalExpression(buffer, context, index + 1);
        var value = evalBinaryOperation(a, b, magicName, typeA, typeB, defaultHandler, buffer, index, context);
        setProperty(target, key, value, allowMagic, precedingNode, operationReferrer, context);
        var valueNode = exports.Utils.createValueNode(value, precedingNode);
        exports.Utils.replaceBuffer(buffer, 0, buffer.length, valueNode);
      };
    };

    var assignmentOperators = [{
      symbol: '=',
      priority: -Infinity,
      ltr: true,
      handler: function handler(buffer, index, context) {
        var operationReferrer = buffer[index];

        if (index === 0) {
          exports.Utils.raise(SyntaxError, 'expect a variable preceding', operationReferrer, context);
        }

        var precedingNode = buffer[index - 1];
        var value = evalExpression(buffer, context, index + 1);

        if (precedingNode.type === 'word') {
          // a = b or a.b = c
          if (index === 1) {
            // a = b
            context.store[precedingNode.value] = value;
          } else {
            // a.b = c
            if (index === 2 || buffer[index - 2].type !== 'symbol' || buffer[index - 2].value !== '.') {
              exports.Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
            }

            var target = evalExpression(buffer, context, 0, index - 2);

            if (!exports.Utils.isDict(target)) {
              exports.Utils.raise(TypeError, 'expect a dict as assignment target', buffer[0], context);
            }

            setProperty(target, precedingNode.value, value, false, buffer[0], operationReferrer, context);
          }
        } else if (precedingNode.type === 'span' && precedingNode.start === '[') {
          // a[b] = c
          var _target = evalExpression(buffer, context, 0, index - 1);

          var key = evalExpression(precedingNode.body, context);
          setProperty(_target, key, value, true, buffer[0], operationReferrer, context);
        } else {
          exports.Utils.raise(SyntaxError, 'invalid assignment', operationReferrer, context);
        }

        var valueNode = exports.Utils.createValueNode(value, operationReferrer);
        exports.Utils.replaceBuffer(buffer, 0, buffer.length, valueNode);
      }
    }, {
      symbol: '+=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__plus', 'number', 'number', function (a, b) {
        return a + b;
      })
    }, {
      symbol: '-=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__minus', 'number', 'number', function (a, b) {
        return a - b;
      })
    }, {
      symbol: '*=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__multiply', 'number', 'number', function (a, b) {
        return a * b;
      })
    }, {
      symbol: '/=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__divide', 'number', 'number', function (a, b) {
        return a / b;
      })
    }, {
      symbol: '%=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__mod', 'number', 'number', function (a, b) {
        return a % b;
      })
    }, {
      symbol: '**=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__pow', 'number', 'number', function (a, b) {
        return Math.pow(a, b);
      })
    }, {
      symbol: '&=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__bitAnd', 'number', 'number', function (a, b) {
        return a & b;
      })
    }, {
      symbol: '^=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__bitXor', 'number', 'number', function (a, b) {
        return a ^ b;
      })
    }, {
      symbol: '|=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__bitOr', 'number', 'number', function (a, b) {
        return a | b;
      })
    }, {
      symbol: '<<=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__leftShift', 'number', 'number', function (a, b) {
        return a << b;
      })
    }, {
      symbol: '>>=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__rightShift', 'number', 'number', function (a, b) {
        return a >> b;
      })
    }, {
      symbol: '>>>=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__unsignedRightShift', 'number', 'number', function (a, b) {
        return a >>> b;
      })
    }, {
      symbol: '&&=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__and', 'boolean', 'boolean', function (a, b) {
        return a && b;
      })
    }, {
      symbol: '||=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__or', 'boolean', 'boolean', function (a, b) {
        return a || b;
      })
    }, {
      symbol: '??=',
      priority: -Infinity,
      ltr: true,
      handler: createAdditionalAssignmentOperator('__nullOr', null, null, function (a, b) {
        return a != null ? a : b;
      })
    }];

    var bitOperators = [{
      symbol: '~',
      priority: 2,
      ltr: false,
      handler: createUnaryOperator('__bitNot', 'number', function (x) {
        return ~x;
      })
    }, {
      symbol: '<<',
      priority: 5,
      ltr: true,
      handler: createBinaryOperator('__leftShift', 'number', 'number', function (a, b) {
        return a << b;
      })
    }, {
      symbol: '>>',
      priority: 5,
      ltr: true,
      handler: createBinaryOperator('__rightShift', 'number', 'number', function (a, b) {
        return a >> b;
      })
    }, {
      symbol: '>>>',
      priority: 5,
      ltr: true,
      handler: createBinaryOperator('__unsignedRightShift', 'number', 'number', function (a, b) {
        return a >>> b;
      })
    }, {
      symbol: '&',
      priority: 8,
      ltr: true,
      handler: createBinaryOperator('__bitAnd', 'number', 'number', function (a, b) {
        return a & b;
      })
    }, {
      symbol: '^',
      priority: 9,
      ltr: true,
      handler: createBinaryOperator('__bitXor', 'number', 'number', function (a, b) {
        return a ^ b;
      })
    }, {
      symbol: '|',
      priority: 10,
      ltr: true,
      handler: createBinaryOperator('__bitOr', 'number', 'number', function (a, b) {
        return a | b;
      })
    }];

    var booleanOperators = [{
      symbol: '!',
      priority: 2,
      ltr: false,
      handler: createUnaryOperator('__not', 'boolean', function (x) {
        return !x;
      })
    }, {
      symbol: '&&',
      priority: 11,
      ltr: true,
      handler: createBinaryOperator('__and', 'boolean', 'boolean', function (a, b) {
        return a && b;
      })
    }, {
      symbol: '||',
      priority: 12,
      ltr: true,
      handler: createBinaryOperator('__or', 'boolean', 'boolean', function (a, b) {
        return a || b;
      })
    }];

    var braceHandler = function braceHandler(buffer, index, context) {
      if (index === 0 || buffer[index - 1].type === 'symbol') {
        // array creation
        var value = evalList(buffer[index].body, context);
        buffer[index] = exports.Utils.createValueNode(value, buffer[index]);
      } else {
        // key access
        var target = evalBufferNode(buffer, index - 1, buffer[index], context);
        var key = evalExpression(buffer[index].body, context);

        var _value = getProperty(target, key, true, buffer[index - 1], buffer[index], context);

        var valueNode = exports.Utils.createValueNode(_value, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 2, valueNode);
      }
    };

    var bracketHandler = function bracketHandler(buffer, index, context) {
      if (index === 0 || buffer[index - 1].type === 'symbol') {
        // dict creation
        var result = Object.create(null);
        var nodes = buffer[index].body;

        if (nodes.length === 0) {
          buffer[index] = exports.Utils.createValueNode(result, buffer[index]);
          return;
        }

        var left = 0;

        for (var right = 0; right <= nodes.length; right++) {
          if (right < nodes.length) {
            var node = nodes[right];

            if (node.type !== 'symbol' || node.value !== ',') {
              continue;
            }
          }

          var firstNode = nodes[left];

          if (firstNode.type === 'symbol' && firstNode.value === '...') {
            // expand
            var props = evalExpression(nodes, context, left + 1, right);

            if (!exports.Utils.isDict(props)) {
              exports.Utils.raise(TypeError, 'expect a dict to expand', firstNode, context);
            }

            Object.assign(result, props);
          } else {
            // insert
            var j = void 0;

            for (j = left; j < right; j++) {
              if (nodes[j].type === 'symbol' && nodes[j].value === '->') {
                break;
              }
            }

            if (j === right) {
              // shortcut
              if (firstNode.type === 'symbol' && firstNode.value === '@') {
                // function
                if (right < left + 3) {
                  exports.Utils.raise(SyntaxError, 'invalid function declaration', firstNode, context);
                }

                var nameNode = nodes[left + 1];

                if (nameNode.type !== 'word') {
                  exports.Utils.raise(SyntaxError, 'invalid function declaration', firstNode, context);
                }

                var name = nameNode.value;
                var func = createInlineFunction(nodes, left + 1, context);
                result[name] = func;
              } else {
                // variable
                var _name = evalExpression(nodes, context, left, right);

                if (typeof _name !== 'string') {
                  exports.Utils.raise(TypeError, 'expect a string as key', firstNode, context);
                }

                if (!(_name in context.store)) {
                  exports.Utils.raise(ReferenceError, exports.Utils.toDisplay(_name) + " is not defined", firstNode, context);
                }

                result[_name] = context.store[_name];
              }
            } else {
              // key -> value
              var _name2 = evalExpression(nodes, context, left, j);

              if (typeof _name2 !== 'string') {
                exports.Utils.raise(TypeError, 'expect a string as key', firstNode, context);
              }

              var value = evalExpression(nodes, context, j + 1, right);
              result[_name2] = value;
            }
          }

          if (right < nodes.length - 1) {
            left = right + 1;
          } else {
            break;
          }
        }

        buffer[index] = exports.Utils.createValueNode(result, buffer[index]);
      } else {
        // callback invocation
        var handler = evalNode(buffer[index - 1], context);

        if (!isInvocable(handler)) {
          exports.Utils.raise(TypeError, 'expect an invocable', buffer[index], context);
        }

        var compiledBody = compileNodes(buffer[index].body, context);

        var callback = function callback(args, referrer, _context) {
          return evalCompiledNodes(compiledBody, _context, true, true);
        };

        var returnValue = invoke(handler, [exports.Utils.createValueNode(callback, buffer[index])], buffer[index], context, null);
        var valueNode = exports.Utils.createValueNode(returnValue, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 2, valueNode);
      }
    };

    var compareOperators = [{
      symbol: '<',
      priority: 6,
      ltr: true,
      handler: createBinaryOperator('__lt', 'number', 'number', function (a, b) {
        return a < b;
      })
    }, {
      symbol: '>',
      priority: 6,
      ltr: true,
      handler: createBinaryOperator('__gt', 'number', 'number', function (a, b) {
        return a > b;
      })
    }, {
      symbol: '<=',
      priority: 6,
      ltr: true,
      handler: createBinaryOperator('__lte', 'number', 'number', function (a, b) {
        return a <= b;
      })
    }, {
      symbol: '>=',
      priority: 6,
      ltr: true,
      handler: createBinaryOperator('__gte', 'number', 'number', function (a, b) {
        return a >= b;
      })
    }, {
      symbol: '==',
      priority: 7,
      ltr: true,
      handler: createBinaryOperator('__equal', null, null, function (a, b) {
        return a === b;
      })
    }, {
      symbol: '!=',
      priority: 7,
      ltr: true,
      handler: createBinaryOperator('__notEqual', null, null, function (a, b) {
        return a !== b;
      })
    }, {
      symbol: '===',
      priority: 7,
      ltr: true,
      handler: function handler(buffer, index, context) {
        var a = evalBufferNode(buffer, index - 1, buffer[index], context);
        var b = evalBufferNode(buffer, index + 1, buffer[index], context);
        var valueNode = exports.Utils.createValueNode(a === b, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
      }
    }, {
      symbol: '!==',
      priority: 7,
      ltr: true,
      handler: function handler(buffer, index, context) {
        var a = evalBufferNode(buffer, index - 1, buffer[index], context);
        var b = evalBufferNode(buffer, index + 1, buffer[index], context);
        var valueNode = exports.Utils.createValueNode(a !== b, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
      }
    }];

    var mathOperators = [{
      symbol: '**',
      priority: 2,
      ltr: false,
      handler: createBinaryOperator('__pow', 'number', 'number', Math.pow)
    }, {
      symbol: '*',
      priority: 3,
      ltr: true,
      handler: createBinaryOperator('__multiply', 'number', 'number', function (a, b) {
        return a * b;
      })
    }, {
      symbol: '/',
      priority: 3,
      ltr: true,
      handler: createBinaryOperator('__divide', 'number', 'number', function (a, b) {
        return a / b;
      })
    }, {
      symbol: '%',
      priority: 3,
      ltr: true,
      handler: createBinaryOperator('__mod', 'number', 'number', function (a, b) {
        return a % b;
      })
    }, {
      symbol: '+',
      priority: 4,
      ltr: true,
      handler: createBinaryOperator('__plus', 'number', 'number', function (a, b) {
        return a + b;
      })
    }, {
      symbol: '-',
      priority: 4,
      ltr: true,
      handler: function handler(buffer, index, context) {
        var b = evalBufferNode(buffer, index + 1, buffer[index], context);

        if (index === 0 || buffer[index - 1].type === 'symbol') {
          // unary
          var result = evalUnaryOperation(b, '__opposite', 'number', function (x) {
            return -x;
          }, buffer, index, context);
          var valueNode = exports.Utils.createValueNode(result, buffer[index]);
          exports.Utils.replaceBuffer(buffer, index, 2, valueNode);
        } else {
          // binary
          var a = evalBufferNode(buffer, index - 1, buffer[index], context);

          var _result = evalBinaryOperation(a, b, '__minus', 'number', 'number', function (a, b) {
            return a - b;
          }, buffer, index, context);

          var _valueNode = exports.Utils.createValueNode(_result, buffer[index]);

          exports.Utils.replaceBuffer(buffer, index - 1, 3, _valueNode);
        }
      }
    }];

    var builtinArray = exports.Utils.injectHelp('A dict providing methods for array manipulations.', exports.Utils.createDict({
      __invoke: exports.Utils.injectHelp('Array.__invoke(data...)', createFunctionHandler(0, Infinity, function (args) {
        return args.slice();
      })),
      create: exports.Utils.injectHelp('Array.create(size = 0, init = null)', createFunctionHandler(0, 2, function (args, referrer, context) {
        if (args.length === 0) {
          return [];
        }

        var count = args[0];

        if (typeof count !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as array size', referrer, context);
        }

        if (count < 0 || !Number.isFinite(count)) {
          exports.Utils.raise(RangeError, 'invalid array size', referrer, context);
        }

        var init = args.length > 1 ? args[1] : null;
        return Array.from({
          length: count
        }, function (_) {
          return init;
        });
      })),
      clone: exports.Utils.injectHelp('Array.clone(array)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to clone', referrer, context);
        }

        return array.slice();
      })),
      sizeOf: exports.Utils.injectHelp('Array.sizeOf(array)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array', referrer, context);
        }

        return array.length;
      })),
      push: exports.Utils.injectHelp('Array.push(array, data...)', createFunctionHandler(2, Infinity, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }

        for (var i = 1; i < args.length; i++) {
          array.push(args[i]);
        }

        return null;
      })),
      unshift: exports.Utils.injectHelp('Array.unshift(array, data...)', createFunctionHandler(2, Infinity, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }

        for (var i = 1; i < args.length; i++) {
          array.unshift(args[i]);
        }

        return null;
      })),
      pop: exports.Utils.injectHelp('Array.pop(array)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }

        return array.length ? array.pop() : null;
      })),
      shift: exports.Utils.injectHelp('Array.shift(array)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to modify', referrer, context);
        }

        return array.length ? array.shift() : null;
      })),
      slice: exports.Utils.injectHelp('Array.slice(array, start = 0, end = Array.sizeOf(array))', createFunctionHandler(1, 3, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as source', referrer, context);
        }

        if (args.length > 1) {
          if (typeof args[1] !== 'number') {
            exports.Utils.raise(TypeError, 'expect a number as begin index', referrer, context);
          }

          if (args.length > 2) {
            if (typeof args[2] !== 'number') {
              exports.Utils.raise(TypeError, 'expect a number as end index', referrer, context);
            }
          }
        }

        return array.slice(args[1], args[2]);
      })),
      insert: exports.Utils.injectHelp('Array.insert(array, index, data...)', createFunctionHandler(3, Infinity, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to operate on', referrer, context);
        }

        var index = args[1];

        if (typeof index !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as start index', referrer, context);
        }

        var arraySize = array.length;
        var normalizedIndex = exports.Utils.normalizeIndex(index, array.length, referrer, context, true);
        var insertCount = args.length - 2;
        array.length += insertCount; // move subsequent elements

        for (var i = arraySize - 1; i >= normalizedIndex; i--) {
          array[i + insertCount] = array[i];
        } // insert new elements


        for (var _i = 2; _i < args.length; _i++) {
          array[normalizedIndex + _i - 2] = args[_i];
        }

        return null;
      })),
      remove: exports.Utils.injectHelp('Array.remove(array, index, count = 1)', createFunctionHandler(2, 3, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var index = args[1];

        if (typeof index !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as start index', referrer, context);
        }

        var arraySize = array.length;
        var normalizedIndex = exports.Utils.normalizeIndex(index, array.length, referrer, context);

        if (args.length === 3) {
          if (typeof args[2] !== 'number') {
            exports.Utils.raise(TypeError, 'expect a number as removing count', referrer, context);
          }

          if (args[2] < 0 || args[2] !== args[2]) {
            exports.Utils.raise(RangeError, 'invalid removing count', referrer, context);
          }
        }

        var count = args.length === 3 ? args[2] : 1;

        if (count === 0) {
          return null;
        }

        if (normalizedIndex + count >= arraySize) {
          array.length = normalizedIndex;
        } else {
          // move trailing elements
          for (var i = normalizedIndex; i < arraySize - count; i++) {
            array[i] = array[i + count];
          }

          array.length -= count;
        }

        return null;
      })),
      clear: exports.Utils.injectHelp('Array.clear(array)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to operate on', referrer, context);
        }

        array.length = 0;
        return null;
      })),
      flat: exports.Utils.injectHelp('Array.flat(arrays, depth = 1)', createFunctionHandler(1, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to flat', referrer, context);
        }

        var depth = args.length === 2 ? args[1] : 1;

        if (typeof depth !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as depth', referrer, context);
        }

        if (depth <= 0 || depth !== depth) {
          exports.Utils.raise(RangeError, 'invalid depth', referrer, context);
        }

        return array.flat(depth);
      })),
      reverse: exports.Utils.injectHelp('Array.reverse(array)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to reverse', referrer, context);
        }

        return array.reverse();
      })),
      unpack: exports.Utils.injectHelp('Array.unpack(array, names, loose = false)', createFunctionHandler(2, 3, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to unpack', referrer, context);
        }

        var names = args[1];

        if (!Array.isArray(names)) {
          exports.Utils.raise(TypeError, 'expect an array of strings as variable names', referrer, context);
        }

        var nameCount = names.length;
        var arraySize = array.length;

        for (var i = 0; i < nameCount; i++) {
          if (typeof names[i] !== 'string') {
            exports.Utils.raise(TypeError, 'expect strings as variable names', referrer, context);
          }
        }

        var loose = args.length === 3 && args[2];

        if (typeof loose !== 'boolean') {
          exports.Utils.raise(TypeError, 'expect a boolean as loose option', referrer, context);
        }

        var store = context.store;

        if (loose) {
          if (nameCount > arraySize) {
            for (var i = 0; i < arraySize; i++) {
              store[names[i]] = array[i];
            }

            for (var _i2 = arraySize; _i2 < nameCount; _i2++) {
              store[names[_i2]] = null;
            }
          } else {
            for (var _i3 = 0; _i3 < nameCount; _i3++) {
              store[names[_i3]] = array[_i3];
            }
          }
        } else {
          if (nameCount > arraySize) {
            exports.Utils.raise(RangeError, 'not enough values in the given array', referrer, context);
          }

          for (var _i4 = 0; _i4 < nameCount; _i4++) {
            store[names[_i4]] = array[_i4];
          }
        }

        return null;
      })),
      indexOf: exports.Utils.injectHelp('Array.indexOf(array, value)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to search in', referrer, context);
        }

        return array.indexOf(args[1]);
      })),
      lastIndexOf: exports.Utils.injectHelp('Array.lastIndexOf(array, value)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to search in', referrer, context);
        }

        return array.lastIndexOf(args[1]);
      })),
      includes: exports.Utils.injectHelp('Array.includes(array, value)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array to check', referrer, context);
        }

        return array.includes(args[1]);
      })),
      sort: exports.Utils.injectHelp('Array.sort(array, compareFn?)', createFunctionHandler(1, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var compareFn = args[1];

        var _compareFn;

        if (args.length > 1 && !isInvocable(compareFn)) {
          exports.Utils.raise(TypeError, 'expect an invocable as compare function', referrer, context);
        }

        if (compareFn) {
          _compareFn = function _compareFn(a, b) {
            return invoke(compareFn, exports.Utils.createRawArray([a, b], referrer), referrer, context, null);
          };
        }

        return array.sort(_compareFn);
      })),
      forEach: exports.Utils.injectHelp('Array.forEach(array, callback)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var callback = args[1];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }

        for (var i = 0; i < array.length; i++) {
          invoke(callback, exports.Utils.createRawArray([array[i], i, array], referrer), referrer, context, null);
        }

        return null;
      })),
      map: exports.Utils.injectHelp('Array.map(array, callback)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var callback = args[1];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }

        return array.map(function (element, index) {
          return invoke(callback, exports.Utils.createRawArray([element, index, array], referrer), referrer, context, null);
        });
      })),
      filter: exports.Utils.injectHelp('Array.filter(array, callback)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var callback = args[1];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }

        return array.filter(function (element, index) {
          var indicator = invoke(callback, exports.Utils.createRawArray([element, index, array], referrer), referrer, context, null);

          if (typeof indicator !== 'boolean') {
            exports.Utils.raise(TypeError, 'expect a boolean as indicator', referrer, context);
          }

          return indicator;
        });
      })),
      every: exports.Utils.injectHelp('Array.every(array, callback)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var callback = args[1];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }

        return array.every(function (element, index) {
          var indicator = invoke(callback, exports.Utils.createRawArray([element, index, array], referrer), referrer, context, null);

          if (typeof indicator !== 'boolean') {
            exports.Utils.raise(TypeError, 'expect a boolean as indicator', referrer, context);
          }

          return indicator;
        });
      })),
      some: exports.Utils.injectHelp('Array.some(array, callback)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var callback = args[1];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }

        return array.some(function (element, index) {
          var indicator = invoke(callback, exports.Utils.createRawArray([element, index, array], referrer), referrer, context, null);

          if (typeof indicator !== 'boolean') {
            exports.Utils.raise(TypeError, 'expect a boolean as indicator', referrer, context);
          }

          return indicator;
        });
      })),
      findIndex: exports.Utils.injectHelp('Array.findIndex(array, callback)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var callback = args[1];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }

        var resultIndex = array.findIndex(function (element, index) {
          var indicator = invoke(callback, exports.Utils.createRawArray([element, index, array], referrer), referrer, context, null);

          if (typeof indicator !== 'boolean') {
            exports.Utils.raise(TypeError, 'expect a boolean as indicator', referrer, context);
          }

          return indicator;
        });

        if (resultIndex === -1) {
          exports.Utils.raise(Error, 'cannot find the wanted element index', referrer, context);
        }

        return resultIndex;
      })),
      find: exports.Utils.injectHelp('Array.find(array, callback)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var array = args[0];

        if (!Array.isArray(array)) {
          exports.Utils.raise(TypeError, 'expect an array as the first argument', referrer, context);
        }

        var callback = args[1];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
        }

        var resultIndex = array.findIndex(function (element, index) {
          var indicator = invoke(callback, exports.Utils.createRawArray([element, index, array], referrer), referrer, context, null);

          if (typeof indicator !== 'boolean') {
            exports.Utils.raise(TypeError, 'expect a boolean as indicator', referrer, context);
          }

          return indicator;
        });

        if (resultIndex === -1) {
          exports.Utils.raise(Error, 'cannot find the wanted element index', referrer, context);
        }

        return array[resultIndex];
      }))
    }));

    var builtinDict = exports.Utils.injectHelp('A dict providing methods for dict manipulations.', exports.Utils.createDict({
      __invoke: exports.Utils.injectHelp('Dict.__invoke(entries?)', createFunctionHandler(0, 1, function (args, referrer, context) {
        var dict = Object.create(null);

        if (args.length === 0) {
          return dict;
        }

        var entries = args[0];

        if (!Array.isArray(entries)) {
          exports.Utils.raise(TypeError, "expect an array of arrays as entries", referrer, context);
        }

        for (var i = 0; i < entries.length; i++) {
          var entry = entries[i];

          if (!Array.isArray(entry) || entry.length !== 2 || typeof entry[0] !== 'string') {
            exports.Utils.raise(TypeError, "invalid dict entry", referrer, context);
          }

          dict[entry[0]] = entry[1];
        }

        return dict;
      })),
      clone: exports.Utils.injectHelp('Dict.clone(dict)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict to clone", referrer, context);
        }

        return exports.Utils.createDict(dict);
      })),
      has: exports.Utils.injectHelp('Dict.has(dict, key)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict to check", referrer, context);
        }

        var key = args[1];

        if (typeof key !== 'string') {
          exports.Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }

        return key in dict;
      })),
      remove: exports.Utils.injectHelp('Dict.remove(dict, key)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict", referrer, context);
        }

        var key = args[1];

        if (typeof key !== 'string') {
          exports.Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }

        if (key in dict) {
          // Don't use `hasProperty` here!
          delete dict[key];
        }

        return null;
      })),
      get: exports.Utils.injectHelp('Dict.get(dict, key)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict", referrer, context);
        }

        var key = args[1];

        if (typeof key !== 'string') {
          exports.Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }

        if (key in dict) {
          // Don't use `hasProperty` here!
          return dict[key];
        } else {
          return null;
        }
      })),
      set: exports.Utils.injectHelp('Dict.set(dict, key, value)', createFunctionHandler(3, 3, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict", referrer, context);
        }

        var key = args[1];

        if (typeof key !== 'string') {
          exports.Utils.raise(TypeError, 'expect a string as key', referrer, context);
        }

        dict[key] = args[2];
        return null;
      })),
      keys: exports.Utils.injectHelp('Dict.keys(dict)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict", referrer, context);
        }

        return Object.keys(dict);
      })),
      entries: exports.Utils.injectHelp('Dict.entries(dict)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict", referrer, context);
        }

        return Object.entries(dict);
      })),
      unpack: exports.Utils.injectHelp('Dict.unpack(dict, names, loose = false)', createFunctionHandler(2, 3, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict to unpack", referrer, context);
        }

        var names = args[1];

        if (!Array.isArray(names)) {
          exports.Utils.raise(TypeError, "expect an array of strings as variable names", referrer, context);
        }

        for (var i = 0; i < names.length; i++) {
          if (typeof names[i] !== 'string') {
            exports.Utils.raise(TypeError, "expect strings as variable names", referrer, context);
          }
        }

        var loose = args.length === 3 && args[2];

        if (typeof loose !== 'boolean') {
          exports.Utils.raise(TypeError, "expect a boolean as loose option", referrer, context);
        }

        var store = context.store;

        if (loose) {
          for (var i = 0; i < names.length; i++) {
            if (hasProperty(dict, names[i], referrer, referrer, context)) {
              store[names[i]] = getProperty(dict, names[i], true, referrer, referrer, context);
            } else {
              store[names[i]] = null;
            }
          }
        } else {
          for (var _i = 0; _i < names.length; _i++) {
            if (!hasProperty(dict, names[_i], referrer, referrer, context)) {
              exports.Utils.raise(ReferenceError, "unknown key(" + exports.Utils.toDisplay(names[_i]) + ")", referrer, context);
            }
          }

          for (var _i2 = 0; _i2 < names.length; _i2++) {
            store[names[_i2]] = getProperty(dict, names[_i2], true, referrer, referrer, context);
          }
        }

        return null;
      })),
      assign: exports.Utils.injectHelp('Dict.assign(dict, props, allowOverwrite = true)', createFunctionHandler(2, 3, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, "expect a dict to modify", referrer, context);
        }

        var props = args[1];

        if (!exports.Utils.isDict(props)) {
          exports.Utils.raise(TypeError, "expect another dict as props", referrer, context);
        }

        var allowOverwrite = args.length === 3 ? args[2] : true;

        if (typeof allowOverwrite !== 'boolean') {
          exports.Utils.raise(TypeError, "expect a boolean as overwrite option", referrer, context);
        }

        getKeysOf(props, referrer, context).forEach(function (key) {
          if (!allowOverwrite && hasProperty(dict, key, referrer, referrer, context)) {
            return;
          }

          var value = getProperty(props, key, true, referrer, referrer, context);
          setProperty(dict, key, value, true, referrer, referrer, context);
        });
        return dict;
      })),
      getConstructorOf: exports.Utils.injectHelp('Dict.getConstructorOf(dict)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, 'expect a dict', referrer, context);
        }

        return getConstructorOf(dict);
      })),
      isInstanceOf: exports.Utils.injectHelp('Dict.isInstanceOf(dict, class)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var dict = args[0];

        if (!exports.Utils.isDict(dict)) {
          exports.Utils.raise(TypeError, 'expect a dict to check', referrer, context);
        }

        return isInstanceOf(args[1], dict);
      }))
    }));

    var builtinFunction = exports.Utils.injectHelp('A dict providing methods for function manipulations.', exports.Utils.createDict({
      isInvocable: exports.Utils.injectHelp('Function.isInvocable(target)', createFunctionHandler(1, 1, function (args, referrer, context) {
        return isInvocable(args[0]);
      })),
      invoke: exports.Utils.injectHelp('Function.invoke(invocable, args = null, thisArg = null)', createFunctionHandler(1, 3, function (args, referrer, context) {
        var fn = args[0];

        if (!isInvocable(fn)) {
          exports.Utils.raise(TypeError, 'expect an invocable to invoke', referrer, context);
        }

        var fnArgs = args.length > 1 ? args[1] : null;

        if (fnArgs !== null && !Array.isArray(fnArgs)) {
          exports.Utils.raise(TypeError, 'expect an array or null as arguments', referrer, context);
        }

        var thisArg = args.length === 3 ? args[2] : null;
        return invoke(fn, fnArgs ? exports.Utils.createRawArray(fnArgs, referrer) : [], referrer, context, thisArg);
      })),
      bind: exports.Utils.injectHelp('Function.bind(invocable, thisArg)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var fn = args[0];

        if (!isInvocable(fn)) {
          exports.Utils.raise(TypeError, 'expect an invocable to bind', referrer, context);
        }

        return function (_args, _referrer, _context) {
          return invoke(fn, _args, _referrer, _context, args[1]);
        };
      }))
    }));

    var builtinNumber = exports.Utils.injectHelp('A dict providing methods for number manipulations.', exports.Utils.createDict({
      toString: exports.Utils.injectHelp('Number.toString(number, radix = 10)', createFunctionHandler(1, 2, function (args, referrer, context) {
        var number = args[0];

        if (typeof number !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number to convert', referrer, context);
        }

        if (args.length === 1) {
          return number.toString();
        }

        var radix = args[1];

        if (typeof radix !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as radix', referrer, context);
        }

        if (!(radix >= 2 && radix <= 36)) {
          exports.Utils.raise(RangeError, 'radix must be between 2 and 36', referrer, context);
        }

        return number.toString(radix);
      })),
      toFixed: exports.Utils.injectHelp('Number.toFixed(number, fractionDigits = 0)', createFunctionHandler(1, 2, function (args, referrer, context) {
        var number = args[0];

        if (typeof number !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number to convert', referrer, context);
        }

        if (args.length === 1) {
          return number.toFixed();
        }

        var fractionDigits = args[1];

        if (typeof fractionDigits !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as count of fraction digits', referrer, context);
        }

        if (!(fractionDigits >= 0 && fractionDigits <= 20)) {
          exports.Utils.raise(RangeError, 'fractionDigits must be between 0 and 20', referrer, context);
        }

        return number.toFixed(fractionDigits);
      })),
      toExponential: exports.Utils.injectHelp('Number.toExponential(number, fractionDigits = 0)', createFunctionHandler(1, 2, function (args, referrer, context) {
        var number = args[0];

        if (typeof number !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number to convert', referrer, context);
        }

        if (args.length === 1) {
          return number.toExponential();
        }

        var fractionDigits = args[1];

        if (typeof fractionDigits !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as count of fraction digits', referrer, context);
        }

        if (!(fractionDigits >= 0 && fractionDigits <= 20)) {
          exports.Utils.raise(RangeError, 'fractionDigits must be between 0 and 20', referrer, context);
        }

        return number.toExponential(fractionDigits);
      })),
      isFinite: exports.Utils.injectHelp('Number.isFinite(number)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var number = args[0];

        if (typeof number !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number', referrer, context);
        }

        return Number.isFinite(number);
      })),
      isInteger: exports.Utils.injectHelp('Number.isInteger(number)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var number = args[0];

        if (typeof number !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number', referrer, context);
        }

        return Number.isInteger(number);
      })),
      isNaN: exports.Utils.injectHelp('Number.isNaN(number)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var number = args[0];

        if (typeof number !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number', referrer, context);
        }

        return Number.isNaN(number);
      })),
      parseInt: exports.Utils.injectHelp('Number.parseInt(string, radix = 10)', createFunctionHandler(1, 2, function (args, referrer, context) {
        var string = args[0];

        if (typeof string !== 'string') {
          exports.Utils.raise(TypeError, 'expect a string to convert', referrer, context);
        }

        var radix = args.length > 1 ? args[1] : 10;

        if (typeof radix !== 'number') {
          exports.Utils.raise(TypeError, 'expect a number as radix', referrer, context);
        }

        if (!(radix >= 2 && radix <= 36)) {
          exports.Utils.raise(RangeError, 'radix must be between 2 and 36', referrer, context);
        }

        var DIGITS = '0123456789abcdefghijklmnopqrstuvwxyz';
        var result = 0;

        for (var i = 0; i < string.length; i++) {
          var x = DIGITS.indexOf(string[string.length - i - 1]);

          if (x === -1 || x >= radix) {
            return NaN;
          }

          result += x * Math.pow(radix, i);
        }

        return result;
      }))
    }));

    var builtinString = exports.Utils.injectHelp('A dict providing methods for string manipulations.', exports.Utils.createDict({
      __invoke: exports.Utils.injectHelp("String.__invoke(parts...)", createFunctionHandler(1, Infinity, function (args, referrer, context) {
        var result = '';

        for (var i = 0; i < args.length; i++) {
          result += exports.Utils.toString(args[i]);
        }

        return result;
      })),
      sizeOf: exports.Utils.injectHelp('String.sizeOf(str)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, 'expect a string', referrer, context);
        }

        return str.length;
      })),
      toLowerCase: exports.Utils.injectHelp("String.toLowerCase(str)", createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'string') {
          exports.Utils.raise(TypeError, "expect a string", referrer, context);
        }

        return args[0].toLowerCase();
      })),
      toUpperCase: exports.Utils.injectHelp("String.toUpperCase(str)", createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'string') {
          exports.Utils.raise(TypeError, "expect a string", referrer, context);
        }

        return args[0].toUpperCase();
      })),
      slice: exports.Utils.injectHelp("String.slice(str, start = 0, end = String.sizeOf(str))", createFunctionHandler(1, 3, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string as source", referrer, context);
        }

        if (args.length > 1) {
          if (typeof args[1] !== 'number') {
            exports.Utils.raise(TypeError, "expect a number as begin index", referrer, context);
          }

          if (args.length > 2) {
            if (typeof args[2] !== 'number') {
              exports.Utils.raise(TypeError, "expect a number as end index", referrer, context);
            }
          }
        }

        return str.slice(args[1], args[2]);
      })),
      indexOf: exports.Utils.injectHelp("String.indexOf(str, substring)", createFunctionHandler(2, 2, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string to search in", referrer, context);
        }

        var substring = args[1];

        if (typeof substring !== 'string') {
          exports.Utils.raise(TypeError, "expect another string to search for", referrer, context);
        }

        return str.indexOf(substring);
      })),
      lastIndexOf: exports.Utils.injectHelp("String.lastIndexOf(str, substring)", createFunctionHandler(2, 2, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string to search in", referrer, context);
        }

        var substring = args[1];

        if (typeof substring !== 'string') {
          exports.Utils.raise(TypeError, "expect another string to search for", referrer, context);
        }

        return str.lastIndexOf(substring);
      })),
      includes: exports.Utils.injectHelp("String.includes(str, substring)", createFunctionHandler(2, 2, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string to search in", referrer, context);
        }

        var substring = args[1];

        if (typeof substring !== 'string') {
          exports.Utils.raise(TypeError, "expect another string to search for", referrer, context);
        }

        return str.includes(substring);
      })),
      repeat: exports.Utils.injectHelp("String.repeat(str, count)", createFunctionHandler(2, 2, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string to repeat", referrer, context);
        }

        var count = args[1];

        if (typeof count !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as repeating count", referrer, context);
        }

        if (!Number.isFinite(count) || count < 0) {
          exports.Utils.raise(RangeError, "invalid count", referrer, context);
        }

        return str.repeat(count);
      })),
      codePointAt: exports.Utils.injectHelp("String.codePointAt(str, index)", createFunctionHandler(2, 2, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string", referrer, context);
        }

        var index = args[1];

        if (typeof index !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as index", referrer, context);
        }

        var normalizedIndex = exports.Utils.normalizeIndex(index, str.length, referrer, context);
        var result = str.codePointAt(normalizedIndex);

        if (typeof result === 'number') {
          return result;
        } else {
          return null;
        }
      })),
      startsWith: exports.Utils.injectHelp("String.startsWith(str, substring)", createFunctionHandler(2, 2, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string to search in", referrer, context);
        }

        var substring = args[1];

        if (typeof substring !== 'string') {
          exports.Utils.raise(TypeError, "expect another string to search for", referrer, context);
        }

        return str.startsWith(substring);
      })),
      endsWith: exports.Utils.injectHelp("String.endsWith(str, substring)", createFunctionHandler(2, 2, function (args, referrer, context) {
        var str = args[0];

        if (typeof str !== 'string') {
          exports.Utils.raise(TypeError, "expect a string to search in", referrer, context);
        }

        var substring = args[1];

        if (typeof substring !== 'string') {
          exports.Utils.raise(TypeError, "expect another string to search for", referrer, context);
        }

        return str.endsWith(substring);
      }))
    }));

    var miscOperators = [{
      symbol: '.',
      priority: 1,
      ltr: true,
      handler: function handler(buffer, index, context) {
        var target = evalBufferNode(buffer, index - 1, buffer[index], context);

        if (!exports.Utils.isDict(target)) {
          exports.Utils.raise(TypeError, 'expect a dict', buffer[index - 1], context);
        }

        var nameNode = buffer[index + 1];

        if (!nameNode || nameNode.type !== 'word') {
          exports.Utils.raise(TypeError, 'expect a word following', buffer[index], context);
        }

        var name = nameNode.value;
        var value = getProperty(target, name, false, buffer[index + 1], buffer[index - 1], context);
        var valueNode = exports.Utils.createValueNode(value, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
      }
    }, {
      symbol: ':',
      priority: 1,
      ltr: true,
      handler: function handler(buffer, index, context) {
        var target = evalBufferNode(buffer, index - 1, buffer[index], context);
        var nameNode = buffer[index + 1];

        if (!nameNode || nameNode.type !== 'word') {
          exports.Utils.raise(TypeError, 'expect a word following', buffer[index], context);
        }

        var dict = null;

        switch (typeof target) {
          case 'number':
            {
              dict = builtinNumber;
              break;
            }

          case 'string':
            {
              dict = builtinString;
              break;
            }

          case 'boolean':
            {
              dict = null;
              break;
            }

          case 'function':
            {
              dict = builtinFunction;
              break;
            }

          case 'object':
            {
              if (!target) {
                dict = null;
              } else if (Array.isArray(target)) {
                dict = builtinArray;
              } else if (exports.Utils.isDict(target)) {
                dict = builtinDict;
              }

              break;
            }

          default:
            {
              dict = null;
            }
        }

        var name = nameNode.value;
        var callback;

        if (dict && name in dict) {
          callback = dict[name];
        } else {
          callback = null;
        }

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable', buffer[index], context);
        }

        var COMMA_NODE = {
          type: 'symbol',
          value: ',',
          line: buffer[index].line,
          column: buffer[index].column,
          offset: buffer[index].offset
        };

        var handler = function handler(rawArgs, referrer, _context, thisArg) {
          return invoke(callback, [exports.Utils.createValueNode(target, referrer), COMMA_NODE].concat(rawArgs), referrer, _context, thisArg);
        };

        var valueNode = exports.Utils.createValueNode(handler, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 3, valueNode);
      }
    }, {
      symbol: '??',
      priority: 12,
      ltr: true,
      handler: createBinaryOperator('__nullOr', null, null, function (a, b) {
        return a != null ? a : b;
      })
    }, {
      symbol: '#',
      priority: 2,
      ltr: true,
      handler: function handler(buffer, index, context) {
        var wordNode = buffer[index + 1];

        if (index + 1 === buffer.length || wordNode.type !== 'word') {
          exports.Utils.raise(SyntaxError, 'expect a word following', buffer[index], context);
        }

        var valueNode = exports.Utils.createValueNode(wordNode.value, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index, 2, valueNode);
      }
    }];

    var parathesisHandler = function parathesisHandler(buffer, index, context) {
      if (index === 0 || buffer[index - 1].type === 'symbol') {
        // pure paratheses
        var value = evalExpression(buffer[index].body, context);
        buffer[index] = exports.Utils.createValueNode(value, buffer[index]);
      } else {
        // function call
        var handler = evalNode(buffer[index - 1], context);

        if (!isInvocable(handler)) {
          exports.Utils.raise(TypeError, 'expect an invocable', buffer[index], context);
        }

        var _value = invoke(handler, buffer[index].body, buffer[index], context, null);

        var valueNode = exports.Utils.createValueNode(_value, buffer[index]);
        exports.Utils.replaceBuffer(buffer, index - 1, 2, valueNode);
      }
    };

    /**
     * List of operator definitions.
     */

    var operators = [].concat(mathOperators, compareOperators, booleanOperators, bitOperators, assignmentOperators, miscOperators, [{
      symbol: '@',
      priority: 0,
      ltr: true,
      handler: inlineFunctionHandler
    }, {
      symbol: '=>',
      priority: 0,
      ltr: true,
      handler: inlineExpressionHandler
    }, {
      symbol: '[',
      priority: 1,
      ltr: true,
      handler: braceHandler
    }, {
      symbol: '(',
      priority: 1,
      ltr: true,
      handler: parathesisHandler
    }, {
      symbol: '{',
      priority: 1,
      ltr: true,
      handler: bracketHandler
    }]);
    /**
     * A utility map. (operator->handler)
     */

    var operatorHandlers = new Map(operators.map(function (op) {
      return [op.symbol, op.handler];
    }));
    /**
     * A utility map. (operator->priority)
     */

    var operatorPriorities = new Map(operators.map(function (op) {
      return [op.symbol, op.priority];
    }));
    /**
     * A utility map. (operator->ltr)
     */

    var operatorLtr = new Map(operators.map(function (op) {
      return [op.symbol, op.ltr];
    }));

    var e = {
      normalSymbols: new Set("()[]{}<>+-*/\\?~!@#$%^&=|,.:;"),
      twoCharacterSymbols: new Set(["&&", "||", "??", "==", ">=", "<=", "!=", "<<", ">>", "++", "--", "**", "+=", "-=", "*=", "/=", "%=", "&=", "^=", "|=", "|>", "->", "=>", "..", "//", "/*", "*/"]),
      threeCharacterSymbols: new Set(["&&=", "||=", "??=", "===", "!==", "<<=", ">>=", "**=", "...", ">>>"]),
      fourCharacterSymbols: new Set([">>>="]),
      globSymbols: new Set(['"', "'", "`"]),
      numberCharacters: new Set("0123456789"),
      extendedNumberCharacters: new Set(".ABCDEF"),
      numberSuffixes: new Set(["D", "B", "O", "H"]),
      spaceCharacters: new Set(" \r\n\t")
    },
        tokenize = function tokenize(t, s) {
      var n = Object.assign({}, e, s),
          a = n.normalSymbols,
          o = n.twoCharacterSymbols,
          l = n.threeCharacterSymbols,
          r = n.fourCharacterSymbols,
          h = n.globSymbols,
          c = n.numberCharacters,
          f = n.extendedNumberCharacters,
          i = n.numberSuffixes,
          u = n.spaceCharacters,
          m = [];
      var p = t[0],
          b = "",
          g = 0,
          S = !1;
      h.has(p) ? b = p : u.has(p) ? g = 1 : c.has(p) ? g = 2 : a.has(p) && (m.push(p), p = "");

      for (var _e = 1; _e < t.length; _e++) {
        var _s = t[_e];

        if (b) {
          p += _s, S ? S = !1 : _s === b ? (m.push(p), p = "", b = "", g = 0) : S = "\\" === _s;
          continue;
        }

        if (h.has(_s)) {
          p && m.push(p), p = _s, b = _s, S = !1;
          continue;
        }

        if ((2 === g || 3 === g) && i.has(_s)) {
          p += _s, g = 3;
          continue;
        }

        if (0 === g && !p && a.has(_s) && m.length) {
          var _e2 = m[m.length - 1];

          if (1 === _e2.length && o.has(_e2 + _s)) {
            m[m.length - 1] += _s;
            continue;
          }

          if (2 === _e2.length && l.has(_e2 + _s)) {
            m[m.length - 1] += _s;
            continue;
          }

          if (3 === _e2.length && r.has(_e2 + _s)) {
            m[m.length - 1] += _s;
            continue;
          }
        }

        var _n = 0;
        if (u.has(_s)) _n = 1;else if (c.has(_s) || 2 === g && f.has(_s) && !p.includes(".")) _n = 2;else if (a.has(_s)) {
          p && m.push(p), m.push(_s), p = "", g = 0;
          continue;
        }
        if (g === _n) p += _s;else if (0 === g && 2 === _n) {
          if (p += _s, p.length > 1) continue;
        } else p && m.push(p), p = _s;
        g = _n;
      }

      return p && m.push(p), m;
    },
        t = Object.assign({}, e, {
      spanSymbols: new Map([["(", ")"], ["[", "]"], ["{", "}"]]),
      escapeCharacters: new Map([["t", "\t"], ["r", "\r"], ["n", "\n"], ["\n", ""], ["\\", "\\"], ['"', '"'], ["'", "'"], ["`", "`"]])
    }),
        token2ast = function token2ast(e, s, n, a, o, l, r) {
      if (n === void 0) {
        n = 0;
      }

      if (a === void 0) {
        a = 1;
      }

      if (o === void 0) {
        o = 1;
      }

      if (l === void 0) {
        l = 0;
      }

      if (r === void 0) {
        r = e.length;
      }

      var h = Object.assign({}, t, s),
          c = h.normalSymbols,
          f = h.twoCharacterSymbols,
          i = h.threeCharacterSymbols,
          u = h.fourCharacterSymbols,
          m = h.globSymbols,
          p = h.numberCharacters,
          b = h.numberSuffixes,
          g = h.spaceCharacters,
          S = h.spanSymbols,
          y = h.escapeCharacters,
          w = [];
      var C = n,
          x = a,
          d = o;

      for (var _t = l; _t < r; _t++) {
        var _s2 = e[_t];

        if (S.has(_s2)) {
          var _n3 = S.get(_s2);

          var _a = 0,
              _o = _t + 1;

          for (; _o < e.length; _o++) {
            if (e[_o] !== _s2) {
              if (e[_o] === _n3) {
                if (!_a) break;
                _a--;
              }
            } else _a++;
          }

          if (_o === e.length) throw new SyntaxError("missing " + _n3 + " for " + _s2 + " at line " + x + " column " + d);
          if (_o === _t + 1) w.push({
            type: "span",
            start: _s2,
            stop: _n3,
            body: [],
            offset: C,
            line: x,
            column: d
          }), C++;else {
            var _a2 = token2ast(e, h, C + 1, x, d + 1, _t + 1, _o),
                _l = _a2.ast;

            w.push({
              type: "span",
              start: _s2,
              stop: _n3,
              body: _l,
              offset: C,
              line: x,
              column: d
            }), C = _a2.stopOffset, x = _a2.stopLine, d = _a2.stopColumn;
          }
          _t = _o;
        } else if (c.has(_s2) || f.has(_s2) || i.has(_s2) || u.has(_s2)) w.push({
          type: "symbol",
          value: _s2,
          offset: C,
          line: x,
          column: d
        });else if (m.has(_s2[0])) {
          var _e3 = _s2[0],
              _t2 = !1;

          for (var _n4 = 1; _n4 < _s2.length - 1; _n4++) {
            var _a3 = _s2[_n4];

            if (_t2) {
              if (!y.has(_a3)) {
                if ("\r" === _a3 && "\n" === _s2[_n4 + 1]) continue;
                throw new SyntaxError("unknown escape character: \\" + _a3);
              }

              _e3 += y.get(_a3), _t2 = !1;
            } else {
              if ("\\" === _a3) {
                _t2 = !0;
                continue;
              }

              _e3 += _a3;
            }
          }

          if (_t2 || _s2[_s2.length - 1] !== _s2[0] || 1 === _s2.length) throw new SyntaxError("missing corresponding " + S + " at line " + x + " column " + d);
          _e3 += _s2[0], w.push({
            type: "glob",
            value: _e3,
            offset: C,
            line: x,
            column: d
          });
        } else if (p.has(_s2[0])) {
          var _e4 = _s2.length - 1;

          for (; _e4 >= 0 && b.has(_s2[_e4]); _e4--) {
          }

          _e4++, w.push({
            type: "number",
            value: _s2.slice(0, _e4),
            suffix: _s2.slice(_e4),
            offset: C,
            line: x,
            column: d
          });
        } else g.has(_s2[0]) || w.push({
          type: "word",
          value: _s2,
          offset: C,
          line: x,
          column: d
        });

        C += _s2.length;

        for (var _e5 = 0; _e5 < _s2.length; _e5++) {
          "\n" === _s2[_e5] && x++;
        }

        var _n2 = _s2.lastIndexOf("\n");

        ~_n2 ? d = _s2.length - _n2 : d += _s2.length;
      }

      return {
        ast: w,
        stopOffset: C,
        stopLine: x,
        stopColumn: d
      };
    },
        parse = function parse(e, t) {
      return token2ast(tokenize(e, t), t);
    };

    var createIf = function createIf(ended) {
      return exports.Utils.injectHelp('if (condition0) { ... } (condition1) { ... } ...', createFunctionHandler(1, 1, function (args, referrer, context) {
        var condition = args[0];

        if (typeof condition !== 'boolean') {
          exports.Utils.raise(TypeError, 'expect a boolean as condition', referrer, context);
        }

        return createFunctionHandler(1, 1, function (_args, referrer, context) {
          var callback = _args[0];

          if (!isInvocable(callback)) {
            exports.Utils.raise(TypeError, 'expect an invocable as callback', referrer, context);
          }

          if (!ended && condition) {
            invoke(callback, [], referrer, context, null);
            return createIf(true);
          } else {
            return createIf(false);
          }
        });
      }));
    };

    var builtinIf = createIf(false);

    var builtinFor = exports.Utils.injectHelp('for (iteratorName, begin, end, step?) { ... }', createFunctionHandler(3, 4, function (args, referrer, context) {
      var iteratorName = args[0];

      if (typeof iteratorName !== 'string') {
        exports.Utils.raise(TypeError, 'expect a string as iterator name', referrer, context);
      }

      var begin = args[1];

      if (!Number.isFinite(begin)) {
        exports.Utils.raise(TypeError, 'expect a finite number as begin value', referrer, context);
      }

      var end = args[2];

      if (!Number.isFinite(end)) {
        exports.Utils.raise(TypeError, 'expect a finite number as end value', referrer, context);
      }

      var defaultStep = begin <= end ? 1 : -1;
      var step = args.length === 4 ? args[3] : defaultStep;

      if (!Number.isFinite(step)) {
        exports.Utils.raise(TypeError, 'expect a finite number as step', referrer, context);
      }

      if (begin !== end && Math.sign(defaultStep) !== Math.sign(step)) {
        exports.Utils.raise(RangeError, 'invalid step', referrer, context);
      }

      return createFunctionHandler(1, 1, function (_args, _referrer, _context) {
        var callback = _args[0];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', _referrer, _context);
        }

        var BREAK_FLAG = Symbol('hxs-break-flag');
        var CONTINUE_FLAG = Symbol('hxs-break-flag');
        var scopeStackLength = _context.stack.length;
        exports.Utils.injectTemp(_context.store, {
          "break": createFunctionHandler(0, 0, function () {
            _context.stack.length = scopeStackLength;
            throw BREAK_FLAG;
          }),
          "continue": createFunctionHandler(0, 0, function () {
            _context.stack.length = scopeStackLength;
            throw CONTINUE_FLAG;
          })
        }, function () {
          if (step > 0) {
            for (var iterator = begin; iterator < end; iterator += step) {
              try {
                _context.store[iteratorName] = iterator;
                invoke(callback, [], referrer, _context, null);
              } catch (error) {
                if (error === BREAK_FLAG) {
                  break;
                } else if (error !== CONTINUE_FLAG) {
                  throw error;
                }
              }
            }
          } else {
            for (var _iterator = begin; _iterator > end; _iterator += step) {
              try {
                _context.store[iteratorName] = _iterator;
                invoke(callback, [], referrer, _context, null);
              } catch (error) {
                if (error === BREAK_FLAG) {
                  break;
                } else if (error !== CONTINUE_FLAG) {
                  throw error;
                }
              }
            }
          }
        });
        return null;
      });
    }));

    var builtinWhile = exports.Utils.injectHelp('while (condition) { ... }', function (rawArgs, referrer, context) {
      return createFunctionHandler(1, 1, function (_args, _referrer, _context) {
        var callback = _args[0];

        if (!isInvocable(callback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as callback', _referrer, _context);
        }

        var BREAK_FLAG = Symbol('hxs-break-flag');
        var CONTINUE_FLAG = Symbol('hxs-break-flag');
        var scopeStackLength = _context.stack.length;
        exports.Utils.injectTemp(_context.store, {
          "break": createFunctionHandler(0, 0, function () {
            _context.stack.length = scopeStackLength;
            throw BREAK_FLAG;
          }),
          "continue": createFunctionHandler(0, 0, function () {
            _context.stack.length = scopeStackLength;
            throw CONTINUE_FLAG;
          })
        }, function () {
          while (true) {
            var condition = evalExpression(rawArgs, context);

            if (typeof condition !== 'boolean') {
              exports.Utils.raise(TypeError, 'expect a boolean as condition', referrer, context);
            }

            if (!condition) {
              break;
            }

            try {
              invoke(callback, [], referrer, _context, null);
            } catch (error) {
              if (error === BREAK_FLAG) {
                break;
              } else if (error !== CONTINUE_FLAG) {
                throw error;
              }
            }
          }
        });
        return null;
      });
    });

    var builtinMath = exports.Utils.injectHelp('A dict providing constants and methods for math calculation.', exports.Utils.createDict({
      PI: Math.PI,
      E: Math.E,
      random: exports.Utils.injectHelp('Math.random()', createFunctionHandler(0, 0, Math.random)),
      sign: exports.Utils.injectHelp('Math.sign(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.sign(args[0]);
      })),
      floor: exports.Utils.injectHelp('Math.floor(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.floor(args[0]);
      })),
      ceil: exports.Utils.injectHelp('Math.ceil(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.ceil(args[0]);
      })),
      round: exports.Utils.injectHelp('Math.round(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.round(args[0]);
      })),
      sqrt: exports.Utils.injectHelp('Math.sqrt(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var x = args[0];

        if (typeof x !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        if (x < 0) {
          exports.Utils.raise(RangeError, "invalid radicant", referrer, context);
        }

        return Math.sqrt(x);
      })),
      min: exports.Utils.injectHelp('Math.min(numbers)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var numbers = args[0];

        if (!Array.isArray(numbers)) {
          exports.Utils.raise(TypeError, "expect an array of numbers", referrer, context);
        }

        var min = Infinity;

        for (var i = 0; i < numbers.length; i++) {
          if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
            exports.Utils.raise(TypeError, "invalid input", referrer, context);
          }

          if (numbers[i] < min) {
            min = numbers[i];
          }
        }

        return min;
      })),
      max: exports.Utils.injectHelp('Math.max(numbers)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var numbers = args[0];

        if (!Array.isArray(numbers)) {
          exports.Utils.raise(TypeError, "expect an array of numbers", referrer, context);
        }

        var max = -Infinity;

        for (var i = 0; i < numbers.length; i++) {
          if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
            exports.Utils.raise(TypeError, "invalid input", referrer, context);
          }

          if (numbers[i] > max) {
            max = numbers[i];
          }
        }

        return max;
      })),
      sum: exports.Utils.injectHelp('Math.sum(numbers)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var numbers = args[0];

        if (!Array.isArray(numbers)) {
          exports.Utils.raise(TypeError, "expect an array of numbers", referrer, context);
        }

        var sum = 0;

        for (var i = 0; i < numbers.length; i++) {
          if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
            exports.Utils.raise(TypeError, "invalid input", referrer, context);
          }

          sum += numbers[i];
        }

        return sum;
      })),
      mean: exports.Utils.injectHelp('Math.mean(numbers)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var numbers = args[0];

        if (!Array.isArray(numbers)) {
          exports.Utils.raise(TypeError, "expect an array of numbers", referrer, context);
        }

        if (!numbers.length) {
          exports.Utils.raise(RangeError, "no numbers given", referrer, context);
        }

        var sum = 0;

        for (var i = 0; i < numbers.length; i++) {
          if (typeof numbers[i] !== 'number' || numbers[i] !== numbers[i]) {
            exports.Utils.raise(TypeError, "invalid input", referrer, context);
          }

          sum += numbers[i];
        }

        return sum / numbers.length;
      })),
      log: exports.Utils.injectHelp('Math.log(a, b)', createFunctionHandler(2, 2, function (args, referrer, context) {
        var a = args[0];

        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as base", referrer, context);
        }

        if (a <= 0 || a === 1 || a === Infinity) {
          exports.Utils.raise(RangeError, "invalid base", referrer, context);
        }

        var b = args[1];

        if (typeof b !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as natural number", referrer, context);
        }

        if (b <= 0) {
          exports.Utils.raise(RangeError, "invalid natural number", referrer, context);
        }

        return Math.log(b) / Math.log(a);
      })),
      ln: exports.Utils.injectHelp('Math.ln(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var x = args[0];

        if (typeof x !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as natural number", referrer, context);
        }

        if (x <= 0) {
          exports.Utils.raise(RangeError, "invalid natural number", referrer, context);
        }

        return Math.log(x);
      })),
      lg: exports.Utils.injectHelp('Math.lg(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var x = args[0];

        if (typeof x !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as natural number", referrer, context);
        }

        if (x <= 0) {
          exports.Utils.raise(RangeError, "invalid natural number", referrer, context);
        }

        return Math.log(x) / Math.LN10;
      })),
      sin: exports.Utils.injectHelp('Math.sin(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.sin(args[0]);
      })),
      cos: exports.Utils.injectHelp('Math.cos(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.cos(args[0]);
      })),
      tan: exports.Utils.injectHelp('Math.tan(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.tan(args[0]);
      })),
      csc: exports.Utils.injectHelp('Math.csc(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return 1 / Math.sin(args[0]);
      })),
      sec: exports.Utils.injectHelp('Math.sec(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return 1 / Math.cos(args[0]);
      })),
      cot: exports.Utils.injectHelp('Math.cot(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return 1 / Math.tan(args[0]);
      })),
      asin: exports.Utils.injectHelp('Math.asin(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.asin(args[0]);
      })),
      acos: exports.Utils.injectHelp('Math.acos(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.acos(args[0]);
      })),
      atan: exports.Utils.injectHelp('Math.atan(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.atan(args[0]);
      })),
      acsc: exports.Utils.injectHelp('Math.acsc(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.asin(1 / args[0]);
      })),
      asec: exports.Utils.injectHelp('Math.asec(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.acos(1 / args[0]);
      })),
      acot: exports.Utils.injectHelp('Math.acot(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.atan(1 / args[0]);
      })),
      sinh: exports.Utils.injectHelp('Math.sinh(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.sinh(args[0]);
      })),
      cosh: exports.Utils.injectHelp('Math.cosh(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.cosh(args[0]);
      })),
      tanh: exports.Utils.injectHelp('Math.tanh(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.tanh(args[0]);
      })),
      csch: exports.Utils.injectHelp('Math.csch(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return 1 / Math.sinh(args[0]);
      })),
      sech: exports.Utils.injectHelp('Math.sech(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return 1 / Math.cosh(args[0]);
      })),
      coth: exports.Utils.injectHelp('Math.coth(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return 1 / Math.tanh(args[0]);
      })),
      asinh: exports.Utils.injectHelp('Math.asinh(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.asinh(args[0]);
      })),
      acosh: exports.Utils.injectHelp('Math.acosh(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.acosh(args[0]);
      })),
      atanh: exports.Utils.injectHelp('Math.atanh(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.atanh(args[0]);
      })),
      acsch: exports.Utils.injectHelp('Math.acsch(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.asinh(1 / args[0]);
      })),
      asech: exports.Utils.injectHelp('Math.asech(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.acosh(1 / args[0]);
      })),
      acoth: exports.Utils.injectHelp('Math.acoth(x)', createFunctionHandler(1, 1, function (args, referrer, context) {
        if (typeof args[0] !== 'number') {
          exports.Utils.raise(TypeError, "expect a number as argument", referrer, context);
        }

        return Math.atanh(1 / args[0]);
      }))
    }));

    var typeOf = exports.Utils.injectHelp('typeOf(value)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var value = args[0];

      switch (typeof value) {
        case 'number':
          {
            return 'number';
          }

        case 'string':
          {
            return 'string';
          }

        case 'boolean':
          {
            return 'boolean';
          }

        case 'function':
          {
            return 'function';
          }

        case 'object':
          {
            if (!value) {
              return 'null';
            } else if (Array.isArray(value)) {
              return 'array';
            } else if (exports.Utils.isDict(value)) {
              return 'dict';
            }
          }

        default:
          {
            exports.Utils.raise(TypeError, 'unrecognized type', referrer, context);
            return null;
          }
      }
    }));
    var number = exports.Utils.injectHelp('number(value)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var type = typeof args[0];

      if (type === 'object' || type === 'function') {
        return NaN;
      } else {
        return Number(args[0]);
      }
    }));
    var string = exports.Utils.injectHelp("string(value, previewSize = 10, previewIndent = '  ')", createFunctionHandler(1, 3, function (args, referrer, context) {
      var previewSize = args.length >= 2 ? args[1] : 10;

      if (typeof previewSize !== 'number') {
        exports.Utils.raise(TypeError, 'expect a number as preview size', referrer, context);
      }

      var previewIndent = args.length >= 3 ? args[2] : '  ';

      if (typeof previewIndent !== 'string') {
        exports.Utils.raise(TypeError, 'expect a string as preview indent', referrer, context);
      }

      return exports.Utils.toString(args[0], previewSize, previewIndent);
    }));

    var _boolean = exports.Utils.injectHelp('boolean(value)', createFunctionHandler(1, 1, function (args, referrer, context) {
      return Boolean(args[0]);
    }));

    var help = exports.Utils.injectHelp('help(target)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var target = args[0];

      switch (typeof target) {
        case 'function':
        case 'object':
          {
            if (target && HELP_SYMBOL in target) {
              return target[HELP_SYMBOL];
            }

            break;
          }
      }

      return 'no help available';
    }));
    var injectHelp = exports.Utils.injectHelp('injectHelp(helpInfo, target)', createFunctionHandler(2, 2, function (args, referrer, context) {
      var helpInfo = args[0];

      if (typeof helpInfo !== 'string') {
        exports.Utils.raise(TypeError, 'expect a string as help info', referrer, context);
      }

      var target = args[1];

      if (typeof target !== 'function' && !exports.Utils.isDict(target)) {
        exports.Utils.raise(TypeError, 'expect a dict or function as target', referrer, context);
      }

      target[HELP_SYMBOL] = helpInfo;
      return target;
    }));

    var exist = exports.Utils.injectHelp('exist(variableName)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var name = args[0];

      if (typeof name !== 'string') {
        exports.Utils.raise(TypeError, 'expect a string as variable name', referrer, context);
      }

      return name in context.store;
    }));
    var delete_ = exports.Utils.injectHelp('delete(variableName)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var name = args[0];

      if (typeof name !== 'string') {
        exports.Utils.raise(TypeError, 'expect a string as variable name', referrer, context);
      }

      var store = context.store;

      if (!(name in store)) {
        exports.Utils.raise(ReferenceError, exports.Utils.toDisplay(name) + " is not defined", referrer, context);
      }

      delete store[name];
      return null;
    }));
    var getContextStore = exports.Utils.injectHelp('getContextStore()', createFunctionHandler(0, 0, function (args, referrer, context) {
      return context.store;
    }));

    var print = exports.Utils.injectHelp('print(message...)', createFunctionHandler(0, Infinity, function (args, referrer, context) {
      var message = '';

      for (var i = 0; i < args.length; i++) {
        message += exports.Utils.toString(args[i]);
      }

      console.log(message);
      return null;
    }));
    var same = exports.Utils.injectHelp('same(a, b)', createFunctionHandler(2, 2, function (args, referrer, context) {
      return Object.is(args[0], args[1]);
    }));
    var keys = exports.Utils.injectHelp('keys(dict)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var dict = args[0];

      if (!exports.Utils.isDict(dict)) {
        exports.Utils.raise(TypeError, 'expect a dict', referrer, context);
      }

      return getKeysOf(dict, referrer, context);
    }));
    var remove = exports.Utils.injectHelp('remove(arrayOrDict, indexOrKey)', createFunctionHandler(2, 2, function (args, referrer, context) {
      removeProperty(args[0], args[1], referrer, referrer, context);
      return null;
    }));
    var sizeOf = exports.Utils.injectHelp('sizeOf(arrayOrString)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var target = args[0];

      if (Array.isArray(target) || typeof target === 'string') {
        return target.length;
      } else {
        exports.Utils.raise(TypeError, "expect an array or string", referrer, context);
        return null; // for type checking
      }
    }));
    var slice = exports.Utils.injectHelp('slice(arrayOrString, begin = 0, end = sizeOf(arrayOrString))', createFunctionHandler(1, 3, function (args, referrer, context) {
      var target = args[0];

      if (Array.isArray(target) || typeof target === 'string') {
        if (args.length > 1) {
          if (typeof args[1] !== 'number') {
            exports.Utils.raise(TypeError, 'expect a number as begin index', referrer, context);
          }

          if (args.length > 2) {
            if (typeof args[2] !== 'number') {
              exports.Utils.raise(TypeError, 'expect a number as end index', referrer, context);
            }
          }
        }

        return target.slice(args[1], args[2]);
      } else {
        exports.Utils.raise(TypeError, "expect an array or string", referrer, context);
        return null; // for type checking
      }
    }));
    var clone = exports.Utils.injectHelp('clone(arrayOrDict)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var target = args[0];

      if (Array.isArray(target)) {
        return target.slice();
      } else if (exports.Utils.isDict(target)) {
        return exports.Utils.createDict(target);
      } else {
        exports.Utils.raise(TypeError, "expect an array or dict to clone", referrer, context);
        return null; // for type checking
      }
    }));
    var indexOf = exports.Utils.injectHelp('indexOf(arrayOrString, valueOrSubstring)', createFunctionHandler(2, 2, function (args, referrer, context) {
      var target = args[0];
      var targetIsString = typeof target === 'string';

      if (!targetIsString && !Array.isArray(target)) {
        exports.Utils.raise(TypeError, 'expect an array to search in', referrer, context);
      }

      if (targetIsString && typeof args[1] !== 'string') {
        exports.Utils.raise(TypeError, 'expect another string to search for', referrer, context);
      }

      return target.indexOf(args[1]);
    }));
    var lastIndexOf = exports.Utils.injectHelp('lastIndexOf(arrayOrString, valueOrSubstring)', createFunctionHandler(2, 2, function (args, referrer, context) {
      var target = args[0];
      var targetIsString = typeof target === 'string';

      if (!targetIsString && !Array.isArray(target)) {
        exports.Utils.raise(TypeError, 'expect an array to search in', referrer, context);
      }

      if (targetIsString && typeof args[1] !== 'string') {
        exports.Utils.raise(TypeError, 'expect another string to search for', referrer, context);
      }

      return target.lastIndexOf(args[1]);
    }));
    var includes = exports.Utils.injectHelp('includes(arrayOrString, valueOrSubstring)', createFunctionHandler(2, 2, function (args, referrer, context) {
      var target = args[0];
      var targetIsString = typeof target === 'string';

      if (!targetIsString && !Array.isArray(target)) {
        exports.Utils.raise(TypeError, 'expect an array to search in', referrer, context);
      }

      if (targetIsString && typeof args[1] !== 'string') {
        exports.Utils.raise(TypeError, 'expect another string to search for', referrer, context);
      }

      return target.includes(args[1]);
    }));

    var builtinRaise = exports.Utils.injectHelp('raise(error_message)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var message = args[0];

      if (typeof message !== 'string') {
        exports.Utils.raise(TypeError, 'expect a string as error message', referrer, context);
      }

      exports.Utils.raise(Error, message, referrer, context);
      return null;
    }));
    var builtinTry = exports.Utils.injectHelp('try { ... } (error) { ... }', createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
      var tryBody = args[0];

      if (!isInvocable(tryBody)) {
        exports.Utils.raise(TypeError, 'expect an invocable to try', referrer, context);
      } // catch


      return function (rawArgs) {
        if (rawArgs.length !== 1 || rawArgs[0].type !== 'word') {
          exports.Utils.raise(TypeError, 'expect a word as the name for error', referrer, context);
        }

        var errName = rawArgs[0].value;
        return createFunctionHandler(1, 1, function (_args, _referrer, _context, _thisArg) {
          var catchBody = _args[0];

          if (!isInvocable(catchBody)) {
            exports.Utils.raise(TypeError, 'expect an invocable as error handler', referrer, context);
          }

          var tryStackLength = context.stack.length;

          try {
            invoke(tryBody, [], referrer, context, thisArg);
          } catch (error) {
            var _Utils$injectTemp;

            if (typeof error === 'symbol') {
              throw error;
            }

            exports.Utils.injectTemp(_context.store, (_Utils$injectTemp = {}, _Utils$injectTemp[errName] = String(error), _Utils$injectTemp), function () {
              context.stack.length = tryStackLength;
              invoke(catchBody, [], _referrer, _context, _thisArg);
            });
          }

          return null;
        });
      };
    }));

    var builtinClass = exports.Utils.injectHelp('A dict providing class APIs.', exports.Utils.createDict({
      __invoke: exports.Utils.injectHelp('Class.__invoke(description)', createFunctionHandler(1, 1, function (args, referrer, context) {
        var description = args[0];

        if (!exports.Utils.isDict(description)) {
          exports.Utils.raise(TypeError, 'expect a dict as class description', referrer, context);
        }

        if ('__init' in description && !isInvocable(description.__init)) {
          exports.Utils.raise(TypeError, 'expect an invocable as initializer', referrer, context);
        }

        return createClass(description, referrer, context);
      })),
      extend: exports.Utils.injectHelp('Class.extend(baseClass, baseArgs = [])(description)', createFunctionHandler(1, 2, function (args, referrer, context) {
        var baseClass = args[0];

        if (!isInvocable(baseClass)) {
          exports.Utils.raise(TypeError, 'expect an invocable as base class', referrer, context);
        }

        var baseArgs = args.length > 1 ? args[1] : [];

        if (!Array.isArray(baseArgs)) {
          exports.Utils.raise(TypeError, 'expect an array as base args', referrer, context);
        }

        return createFunctionHandler(1, 1, function (_args, _referrer, _context) {
          var description = _args[0];

          if (!exports.Utils.isDict(description)) {
            exports.Utils.raise(TypeError, 'expect a dict as class description', referrer, context);
          }

          if ('__init' in description && !isInvocable(description.__init)) {
            exports.Utils.raise(TypeError, 'expect an invocable as initializer', referrer, context);
          }

          if ('__init' in description) {
            var oldInit = description.__init;

            var newInit = function newInit(rawArgs, ref, ctx, thisArg) {
              invoke(baseClass, exports.Utils.createRawArray(baseArgs, ref), referrer, context, thisArg);
              return invoke(oldInit, rawArgs, ref, ctx, thisArg);
            };

            description.__init = newInit;
          } else {
            var init = function init(rawArgs, ref, ctx, thisArg) {
              invoke(baseClass, exports.Utils.createRawArray(baseArgs, ref), referrer, context, thisArg);
              return null;
            };

            description.__init = init;
          }

          var constructor = createClass(description, _referrer, _context);
          constructor[BASE_SYMBOL] = baseClass;
          return constructor;
        });
      }))
    }));

    var getExports = exports.Utils.injectHelp('getExports()', createFunctionHandler(0, 0, function (args, referrer, context) {
      return context.exports;
    }));
    var export_ = exports.Utils.injectHelp('export(exports)', createFunctionHandler(1, 1, function (args, referrer, context) {
      var exports$1 = args[0];

      if (!exports.Utils.isDict(exports$1)) {
        exports.Utils.raise(TypeError, "expect a dict as exports", referrer, context);
      }

      Object.assign(context.exports, exports$1);
      return null;
    }));

    var builtinJSON = exports.Utils.injectHelp('A dict providing methods for JSON operations.', exports.Utils.createDict({
      parse: exports.Utils.injectHelp("JSON.parse(string)", createFunctionHandler(1, 1, function (args, referrer, context) {
        var string = args[0];

        if (typeof string !== 'string') {
          exports.Utils.raise(TypeError, "expect a string to parse", referrer, context);
        }

        return JSON.parse(string);
      })),
      stringify: exports.Utils.injectHelp("JSON.stringify(value, space = '')", createFunctionHandler(1, 2, function (args, referrer, context) {
        var space = args.length > 1 ? args[1] : '';

        if (typeof space !== 'string') {
          exports.Utils.raise(TypeError, "expect a string as space argument", referrer, context);
        }

        return JSON.stringify(args[0], undefined, space);
      }))
    }));

    var builtinPromise = exports.Utils.injectHelp('A dict providing APIs of promises.', createClass({
      __init: createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
        var initializer = args[0];

        if (!isInvocable(initializer)) {
          exports.Utils.raise(TypeError, 'expect an invocable as promise initializer', referrer, context);
        }

        var promise = new Promise(function (resolve, reject) {
          var _resolve = createFunctionHandler(1, 1, function (_args, _referrer, _context, _thisArg) {
            resolve(_args[0]);
            return null;
          });

          var _reject = createFunctionHandler(1, 1, function (_args, _referrer, _context, _thisArg) {
            reject(_args[0]);
            return null;
          });

          invoke(initializer, exports.Utils.createRawArray([_resolve, _reject], referrer), referrer, context, null);
        });
        thisArg[PROMISE_SYMBOL] = promise;
        return null;
      }),
      then: createFunctionHandler(1, 2, function (args, referrer, context, thisArg) {
        var resolveCallback = args[0];

        if (!isInvocable(resolveCallback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as resolve callback', referrer, context);
        }

        var rejectCallback = args[1];

        if (args.length > 1 && !isInvocable(rejectCallback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as reject callback', referrer, context);
        }

        var _resolveCallback = function _resolveCallback(data) {
          var result = invoke(resolveCallback, exports.Utils.createRawArray([exports.Utils.filterValue(data)], referrer), referrer, context, null);

          if (exports.Utils.isDict(result) && isInstanceOf(builtinPromise, result)) {
            return result[PROMISE_SYMBOL];
          } else {
            return result;
          }
        };

        var _rejectCallback = function _rejectCallback(reason) {
          invoke(rejectCallback, exports.Utils.createRawArray([exports.Utils.filterValue(reason)], referrer), referrer, context, null);
        };

        var thisPromise = thisArg[PROMISE_SYMBOL];
        var nextPromise = invoke(builtinPromise, exports.Utils.createRawArray([function () {
          return null;
        }], referrer), referrer, context, null);
        nextPromise[PROMISE_SYMBOL] = thisPromise.then(_resolveCallback, _rejectCallback);
        return nextPromise;
      }),
      "catch": createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
        var rejectCallback = args[0];

        if (!isInvocable(rejectCallback)) {
          exports.Utils.raise(TypeError, 'expect an invocable as reject callback', referrer, context);
        }

        var _rejectCallback = function _rejectCallback(reason) {
          invoke(rejectCallback, exports.Utils.createRawArray([exports.Utils.filterValue(reason)], referrer), referrer, context, null);
        };

        var thisPromise = thisArg[PROMISE_SYMBOL];
        var nextPromise = invoke(builtinPromise, exports.Utils.createRawArray([function () {
          return null;
        }], referrer), referrer, context, null);
        nextPromise[PROMISE_SYMBOL] = thisPromise["catch"](_rejectCallback);
        return nextPromise;
      })
    }, null, null));
    Object.assign(builtinPromise, {
      resolve: exports.Utils.injectHelp('Promise.resolve(data)', createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
        var initializer = createFunctionHandler(2, 2, function (_args, _referrer, _context, _thisArg) {
          invoke(_args[0], // resolve
          exports.Utils.createRawArray([args[0]], referrer), referrer, context, null);
          return null;
        });
        return invoke(builtinPromise, exports.Utils.createRawArray([initializer], referrer), referrer, context, thisArg);
      })),
      reject: exports.Utils.injectHelp('Promise.reject(reason)', createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
        var initializer = createFunctionHandler(2, 2, function (_args, _referrer, _context, _thisArg) {
          invoke(_args[1], // reject
          exports.Utils.createRawArray([args[0]], referrer), referrer, context, null);
          return null;
        });
        return invoke(builtinPromise, exports.Utils.createRawArray([initializer], referrer), referrer, context, thisArg);
      }))
    });
    Object.assign(builtinPromise, {
      all: exports.Utils.injectHelp('Promise.all(promises)', createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
        var promises = args[0];

        if (!Array.isArray(promises) || !promises.every(function (promise) {
          return exports.Utils.isDict(promise) && isInstanceOf(builtinPromise, promise);
        })) {
          exports.Utils.raise(TypeError, 'expect an array of promises', referrer, context);
        }

        var resultPromise = invoke(builtinPromise, exports.Utils.createRawArray([function () {
          return null;
        }], referrer), referrer, context, null);
        resultPromise[PROMISE_SYMBOL] = Promise.all(promises.map(function (promise) {
          return promise[PROMISE_SYMBOL];
        }));
        return resultPromise;
      })),
      any: exports.Utils.injectHelp('Promise.any(promises)', createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
        var promises = args[0];

        if (!Array.isArray(promises) || !promises.every(function (promise) {
          return exports.Utils.isDict(promise) && isInstanceOf(builtinPromise, promise);
        })) {
          exports.Utils.raise(TypeError, 'expect an array of promises', referrer, context);
        }

        var resultPromise = invoke(builtinPromise, exports.Utils.createRawArray([function () {
          return null;
        }], referrer), referrer, context, null);
        resultPromise[PROMISE_SYMBOL] = Promise.any(promises.map(function (promise) {
          return promise[PROMISE_SYMBOL];
        }));
        return resultPromise;
      })),
      race: exports.Utils.injectHelp('Promise.race(promises)', createFunctionHandler(1, 1, function (args, referrer, context, thisArg) {
        var promises = args[0];

        if (!Array.isArray(promises) || !promises.every(function (promise) {
          return exports.Utils.isDict(promise) && isInstanceOf(builtinPromise, promise);
        })) {
          exports.Utils.raise(TypeError, 'expect an array of promises', referrer, context);
        }

        var resultPromise = invoke(builtinPromise, exports.Utils.createRawArray([function () {
          return null;
        }], referrer), referrer, context, null);
        resultPromise[PROMISE_SYMBOL] = Promise.race(promises.map(function (promise) {
          return promise[PROMISE_SYMBOL];
        }));
        return resultPromise;
      }))
    });

    var builtins = exports.Utils.createDict({
      "true": true,
      "false": false,
      Infinity: Infinity,
      NaN: NaN,
      "null": null,
      _: null,
      "if": builtinIf,
      "while": builtinWhile,
      "for": builtinFor,
      raise: builtinRaise,
      "try": builtinTry,
      Class: builtinClass,
      Array: builtinArray,
      String: builtinString,
      Dict: builtinDict,
      Function: builtinFunction,
      Number: builtinNumber,
      Math: builtinMath,
      JSON: builtinJSON,
      Promise: builtinPromise,
      help: help,
      injectHelp: injectHelp,
      exist: exist,
      "delete": delete_,
      getContextStore: getContextStore,
      typeOf: typeOf,
      number: number,
      string: string,
      "boolean": _boolean,
      print: print,
      same: same,
      keys: keys,
      remove: remove,
      sizeOf: sizeOf,
      slice: slice,
      clone: clone,
      indexOf: indexOf,
      lastIndexOf: lastIndexOf,
      includes: includes,
      getExports: getExports,
      "export": export_,
      forEach: builtinArray.forEach,
      map: builtinArray.map,
      filter: builtinArray.filter,
      assign: builtinDict.assign,
      getConstructorOf: builtinDict.getConstructorOf,
      isInstanceOf: builtinDict.isInstanceOf,
      isInvocable: builtinFunction.isInvocable,
      invoke: builtinFunction.invoke,
      bind: builtinFunction.bind,
      isNaN: builtinNumber.isNaN
    });

    /**
     * Execute the given code.
     */

    var evalCode = function evalCode(code, context) {
      var _context = {};

      if (context) {
        Object.assign(_context, context);
      }

      if (!_context.store) {
        _context.store = exports.Utils.createDict(builtins);
      }

      if (!_context.exports) {
        _context.exports = Object.create(null);
      }

      if (!_context.resolvedModules) {
        _context.resolvedModules = Object.create(null);
      }

      if (!_context.source) {
        _context.source = 'unknown';
      }

      if (!_context.basePath) {
        _context.basePath = '';
      }

      if (!_context.stack) {
        _context.stack = [];
      }

      return evalNodes(parse(code).ast, _context);
    };

    exports.BASE_SYMBOL = BASE_SYMBOL;
    exports.CONSTRUCTOR_SYMBOL = CONSTRUCTOR_SYMBOL;
    exports.HELP_SYMBOL = HELP_SYMBOL;
    exports.PROMISE_SYMBOL = PROMISE_SYMBOL;
    exports.builtins = builtins;
    exports.compileExpression = compileExpression;
    exports.compileNodes = compileNodes;
    exports.createAdditionalAssignmentOperator = createAdditionalAssignmentOperator;
    exports.createBinaryOperator = createBinaryOperator;
    exports.createClass = createClass;
    exports.createFunctionHandler = createFunctionHandler;
    exports.createInlineExpression = createInlineExpression;
    exports.createInlineFunction = createInlineFunction;
    exports.createUnaryOperator = createUnaryOperator;
    exports.evalBinaryOperation = evalBinaryOperation;
    exports.evalBufferNode = evalBufferNode;
    exports.evalCode = evalCode;
    exports.evalCompiledExpression = evalCompiledExpression;
    exports.evalCompiledNodes = evalCompiledNodes;
    exports.evalExpression = evalExpression;
    exports.evalList = evalList;
    exports.evalNode = evalNode;
    exports.evalNodes = evalNodes;
    exports.evalUnaryOperation = evalUnaryOperation;
    exports.getConstructorOf = getConstructorOf;
    exports.getKeysOf = getKeysOf;
    exports.getOperatorSymbol = getOperatorSymbol;
    exports.getProperty = getProperty;
    exports.hasProperty = hasProperty;
    exports.inlineExpressionHandler = inlineExpressionHandler;
    exports.inlineFunctionHandler = inlineFunctionHandler;
    exports.invoke = invoke;
    exports.isInstanceOf = isInstanceOf;
    exports.isInvocable = isInvocable;
    exports.operatorHandlers = operatorHandlers;
    exports.operatorLtr = operatorLtr;
    exports.operatorPriorities = operatorPriorities;
    exports.operators = operators;
    exports.parseArg = parseArg;
    exports.parseArgList = parseArgList;
    exports.removeProperty = removeProperty;
    exports.setProperty = setProperty;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
